"""
Sphere obstacle description
"""
from parmepy.constants import np, PARMES_INTEGER
from parmepy.domain.obstacle.obstacle_3D import Obstacle3D


class Sphere(Obstacle3D):
    """
    Discrete obstacles representation.
    """

    def __init__(self, obst, radius=0.2):
        """
        Creates the sphere obsctacle and z-boundaries and 
        returns 3 arrays corresponding to the characteristic 
        functions of three different porous media.

        @param obst : Three dimensional parent obstacle.
        @param radius : Sphere radius
        """
        ## 3D parent obstacle
        self.obst = obst
        ## Radius of the sphere
        self.radius = radius
        ## Characteristic function (array) for z-boundaries
        self.chiBoundary = None
        ## Characteristic function (array) for the solid
        self.chiSolid = None
        ## Characteristic function (array) for the porous area
        self.chiPorous = None
        print 'obstacle =', self.obst.obstacleName

    def setUp(self, topology):
        """
        Compute the characteristic functions associated 
        to z-boundaries and sphere
        """
        ## Temporary arrays
        chiBoundary_i = []
        chiBoundary_j = []
        chiBoundary_k = []
        chiSolid_i = []
        chiSolid_j = []
        chiSolid_k = []
        chiPorous_i = []
        chiPorous_j = []
        chiPorous_k = []

        step = topology.mesh.space_step
        ghosts = topology.ghosts
        local_start = topology.mesh.local_start
        local_end = topology.mesh.local_end
        coord_start = topology.mesh.origin + (ghosts * step)
        coord_end = topology.mesh.end - (ghosts * step)
        layerMin = coord_start[2] + self.obst.zlayer
        layerMax = coord_end[2] - self.obst.zlayer
        if not (self.obst.porousLayerThickn <= self.radius):
            raise ValueError("Error, porous layer thickness" +
                             "is higher than sphere radius.")
        radiusMinuslayer = self.radius- self.obst.porousLayerThickn

        print 'step, ghosts, local_start, local_end, zlayer', \
              step, ghosts, local_start, local_end, self.obst.zlayer
        print 'start, end, layerMin, layerMax, radiusMinuslayer', \
              coord_start, coord_end, layerMin, layerMax, radiusMinuslayer

        for k in xrange (local_start[2], local_end[2] + 1):
            cz = coord_start[2] + k * step[2]
            for j in xrange (local_start[1], local_end[1] + 1):
                cy = coord_start[1] + j * step[1]
                for i in xrange (local_start[0], local_end[0] + 1):
                    if (cz >= layerMax or cz <= layerMin):
                    # we are in the z-layer boundary:
                        chiBoundary_i.append(i)
                        chiBoundary_j.append(j)
                        chiBoundary_k.append(k)
                    else :
                        cx = coord_start[0] + i * step[0]
                        dist = np.sqrt((cx - self.obst.center[0]) ** 2 +
                               (cy - self.obst.center[1]) ** 2 +
                               (cz - self.obst.center[2]) ** 2)
                        if (radiusMinuslayer < dist 
                            and dist <= self.radius + 1E-12 
                            and self.obst.porousLayerThickn != 0.):
                            # we are in the porous region of the sphere:
                            chiPorous_i.append(i)
                            chiPorous_j.append(j)
                            chiPorous_k.append(k)
                        if (dist <= radiusMinuslayer + 1E-12):
                            # we are in the solid region of the sphere:
                            chiSolid_i.append(i)
                            chiSolid_j.append(j)
                            chiSolid_k.append(k)

        ## Characteristic function of penalized boundaries
        chiBoundary_i = np.asarray(chiBoundary_i, dtype=PARMES_INTEGER)
        chiBoundary_j = np.asarray(chiBoundary_j, dtype=PARMES_INTEGER)
        chiBoundary_k = np.asarray(chiBoundary_k, dtype=PARMES_INTEGER)
        self.chiBoundary = tuple([chiBoundary_i, chiBoundary_j, chiBoundary_k])

        ## Characteristic function of solid areas
        chiSolid_i = np.asarray(chiSolid_i, dtype=PARMES_INTEGER)
        chiSolid_j = np.asarray(chiSolid_j, dtype=PARMES_INTEGER)
        chiSolid_k = np.asarray(chiSolid_k, dtype=PARMES_INTEGER)
        self.chiSolid = tuple([chiSolid_i, chiSolid_j, chiSolid_k])

        ## Characteristic function of porous areas
        chiPorous_i = np.asarray(chiPorous_i, dtype=PARMES_INTEGER)
        chiPorous_j = np.asarray(chiPorous_j, dtype=PARMES_INTEGER)
        chiPorous_k = np.asarray(chiPorous_k, dtype=PARMES_INTEGER)
        self.chiPorous = tuple([chiPorous_i, chiPorous_j, chiPorous_k])

    def __str__(self):
        """ToString method"""
        return "Sphere"

if __name__ == "__main__" :
    print "This module defines the following classe:"
    print "Sphere: ", Sphere.__doc__
