"""
@package parmepy.fields.continuous

Continuous variable description
"""
from parmepy import __VERBOSE__
from parmepy.constants import np, PARMES_INTEGER, debug
from parmepy.fields.scalar import ScalarField
from parmepy.fields.vector import VectorField


class Field(object):
    """ A variable defined on a physical domain """

    @debug
    def __new__(cls, *args, **kw):
        return object.__new__(cls, *args, **kw)

    @debug
    def __init__(self, domain, name="?", vector=False):
        """
        Create a Field.

        @param domain : variable application domain.
        @param name : name of the variable (used for vtk output).
        @param vector : is variable is a vector.
        """
        ## Application domain of the variable.
        self.domain = domain
        ## Field dimension.
        self.dimension = self.domain.dimension
        ## Discretisation of this field
        self.discreteField = None
        ## Name of this field
        self.name = name
        # Number of different discretizations
        self._fieldId = -1
        ## List of the various discretizations of this field
        self.discreteField = []
        ## List of the various topologies that discretizations are based on
        self.topologies = []
        ## Is field is a vector field
        self.vector = vector

    @debug
    def discretize(self, topo):
        """
        Discretization of the field on a topology.
        @param topo : topology specification for discretization
        @return discrete field and its index.

        """
        # Find if topology already use to discretize self
        try:
            fieldId = self.topologies.index(topo)
            #print "Already discretized with topology:"
            # topo already used
        except ValueError:
            # Create a new discrete field with topo
            self._fieldId += 1
            fieldId = self._fieldId
            if self.vector:
                self.discreteField.insert(
                    fieldId, VectorField(self, topo,
                                         name=self.name + "_D_" + str(fieldId),
                                         idFromParent=fieldId))
            else:
                self.discreteField.insert(
                    fieldId, ScalarField(self, topo,
                                         name=self.name + "_D_" + str(fieldId),
                                         idFromParent=fieldId))
            self.topologies.insert(fieldId, topo)
            #print "Discretized with topology:"
        #print self.topologies[fieldId]
        return self.discreteField[fieldId], fieldId

    @debug
    def initialize(self):
        """
        Initialize a field.
        @see VectorField.initialize
        @see ScalarField.initialize
        """
        if self._fieldId == -1:
            raise ValueError("Cannot initialise analytical" +
                             " field non discretized.")
        if __VERBOSE__:
            print self.name,
        for dF in self.discreteField:
            dF.initialize()
        if __VERBOSE__:
            print " .Done"

    def __str__(self):
        """ Field info display """
        s = self.name + ' , ' + str(self.dimension) + 'D field '
        if len(self.discreteField) != 0:
            s += "with the following discretizations:\n"
            for i in range(len(self.discreteField)):
                s += self.discreteField[i].__str__()
        else:
            s += ". Not discretised\n"
        return s


if __name__ == "__main__":
    print __doc__
    print "- Provided class : ContinuousField"
    print ContinuousField.__doc__
