"""
@package parmepy.fields.gpu_discrete
"""
from parmepy import __VERBOSE__
from parmepy.constants import ORDER, np, debug
from parmepy.gpu import cl, PARMES_REAL_GPU
from parmepy.fields.vector import VectorField
from parmepy.fields.scalar import ScalarField
from abc import ABCMeta, abstractmethod


class GPUDiscreteField(object):

    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self, queue, precision):
        self.queue = queue
        self.gpu_data = None
        self.initialization_kernel = None
        self.precision = precision
        self.mem_size = 0

    @abstractmethod
    def toHost(self):
        pass

    @abstractmethod
    def toDevice(self):
        pass


class GPUVectorField(VectorField, GPUDiscreteField):
    """
    GPU Discrete field implementation
    """
    def __init__(self, queue, parent,
                 topology=None, name="?", idFromParent=None,
                 precision=PARMES_REAL_GPU):
        super(GPUVectorField, self).__init__(self, queue, precision,
                                             parent, topology,
                                             name, idFromParent)
        self.gpu_data = [None for d in xrange(self.dimension)]
        for d in xrange(self.dimension):
            self.data[d] = np.asarray(self.data[d],
                                      dtype=precision, order=ORDER)
            self.gpu_data[d] = cl.Buffer(self.queue.context,
                                         cl.mem_flags.READ_WRITE,
                                         size=self.data[d].nbytes)
            self.mem_size += self.gpu_data[d].size
        if __VERBOSE__:
            print self.name, self.mem_size, "Bytes (",
            print self.mem_size / (1024 ** 2), "MB)"
        else:
            print ".",

    @classmethod
    def fromVectorField(cls, queue, vfield, precision=PARMES_REAL_GPU):
        vfield.__class__ = cls
        GPUDiscreteField.__init__(vfield, queue, precision)
        vfield.gpu_data = [None for d in xrange(vfield.dimension)]
        for d in xrange(vfield.dimension):
            vfield.data[d] = np.asarray(vfield.data[d],
                                        dtype=precision, order=ORDER)
            vfield.gpu_data[d] = cl.Buffer(vfield.queue.context,
                                           cl.mem_flags.READ_WRITE,
                                           size=vfield.data[d].nbytes)
            vfield.mem_size += vfield.gpu_data[d].size
        if __VERBOSE__:
            print vfield.name, vfield.mem_size, "Bytes (",
            print vfield.mem_size / (1024 ** 2), "MB)"
        else:
            print ".",

    @debug
    def initialize(self, formula=None):
        if self.initialization_kernel is None:
            VectorField.initialize(self, formula)
            if self.dimension == 2:
                self.data[0] = np.asarray(self.data[0],
                                          dtype=self.precision, order=ORDER)
                self.data[1] = np.asarray(self.data[1],
                                          dtype=self.precision, order=ORDER)
            else:
                self.data[0] = np.asarray(self.data[0],
                                          dtype=self.precision, order=ORDER)
                self.data[1] = np.asarray(self.data[1],
                                          dtype=self.precision, order=ORDER)
                self.data[2] = np.asarray(self.data[2],
                                          dtype=self.precision, order=ORDER)
            data, time = self.toDevice()
            print "Transfers : ", self.mem_size, "Bytes transfered at ",
            print "{0:.3f} GBytes/sec".format(
                (self.mem_size * 1e-9) / time)
        else:
            coord_min = np.ones(4, dtype=self.precision)
            mesh_size = np.ones(4, dtype=self.precision)
            coord_min[0:self.dimension] = np.asarray(self.topology.mesh.origin,
                                                     dtype=self.precision)
            mesh_size[0:self.dimension] = np.asarray(self.topology.mesh.space_step,
                                                     dtype=self.precision)
            if self.dimension == 2:
                evt = self.initialization_kernel(self.gpu_data[0],
                                                 self.gpu_data[1],
                                                 coord_min,
                                                 mesh_size)
            else:
                evt = self.initialization_kernel(self.gpu_data[0],
                                                 self.gpu_data[1],
                                                 self.gpu_data[2],
                                                 coord_min,
                                                 mesh_size)
            self.queue.finish()
            cTime = (evt.profile.end - evt.profile.start) * 1e-9
            print "Computing  : ", cTime, "sec"
        self.contains_data = False
        self.data_on_device = True

    def toDevice(self):
        """
        Host to device method.

        @param queue : OpenCL queue.

        Performs a direct OpenCL copy from numpy arrays
        to OpenCL Buffers.\n
        Arrange memory on device so that vector components are
        contiguous in the direction of the component.\n
        Example : A 3D vector field F(x,y,z) is made up of 3
        OpenCL Buffers Fx, Fy, Fz. The memory layout is :
        - Fx : x-major ordering. On device,
        Fx[i + j*WIDTH + k*WIDTH*HEIGHT] access to Fx(i,j,k)
        - Fy : y-major ordering. On device,
        Fy[i + j*WIDTH + k*WIDTH*HEIGHT] access to Fy(j,i,k)
        - Fz : z-major ordering. On device,
        Fz[i + j*WIDTH + k*WIDTH*HEIGHT] access to Fz(k,i,j)
        """
        print "host->device :", self.name,
        time = 0.
        evt = [None, None, None]
        evt[0] = cl.enqueue_copy(
            self.queue, self.gpu_data[0],
            self.data[0].ravel(order=ORDER))
        evt[1] = cl.enqueue_copy(
            self.queue, self.gpu_data[1],
            self.data[1].swapaxes(0, 1).ravel(order=ORDER))
        if self.dimension == 3:
            evt[2] = cl.enqueue_copy(
                self.queue, self.gpu_data[2],
                self.data[2].swapaxes(0, 1).swapaxes(0, 2).ravel(order=ORDER))
        self.queue.finish()
        for e in evt:
            if e is not None:
                time += (e.profile.end - e.profile.start) * 1e-9
        if __VERBOSE__:
            print self.mem_size, "Bytes transfered at ",
            print "{0:.3f} GBytes/sec".format(
                self.mem_size / (time * 1024 ** 3))
        else:
            print "Transfer host->device"
        return self.mem_size, time

    def toHost(self):
        """
        Device to host method.

        @param queue : OpenCL queue.
        @param discreteField : Variable to transfer.

        Performs a direct OpenCL copy from OpenCL Buffers
        to numpy arrays.\n
        As memory layout is arranged on device, not only a
        copy is performed but also transpositions to have numpy
        arrays consistent to each other.
        """
        if not self.contains_data:
            print "device->host :", self.name,
            time = 0.
            evt = [None, None, None]
            # Use of discreteField.data[XDIR] as temporary array
            #\todo Need cleaner solution
            # Get ZDIR component
            if self.dimension == 3:
                evt[2] = cl.enqueue_copy(self.queue, self.data[0],
                                         self.gpu_data[2])
                self.data[2] = self.data[0]\
                    .swapaxes(0, 2).swapaxes(0, 1)
            # Get YDIR component
            evt[1] = cl.enqueue_copy(self.queue, self.data[0],
                                     self.gpu_data[1])
            self.data[1] = self.data[0].swapaxes(0, 1)
            # Get XDIR component
            evt[0] = cl.enqueue_copy(self.queue, self.data[0],
                                     self.gpu_data[0])
            self.queue.finish()
            for e in evt:
                if e is not None:
                    time += (e.profile.end - e.profile.start) * 1e-9
            if __VERBOSE__:
                print self.mem_size, "Bytes transfered at ",
                print "{0:.3f} GBytes/sec".format(
                    self.mem_size / (time * 1024 ** 3))
                return self.mem_size, time
            else:
                print 'Transfer device->host'


class GPUScalarField(ScalarField, GPUDiscreteField):
    """
    GPU Discrete field implementation
    """
    def __init__(self, queue, parent,
                 topology=None, name="?", idFromParent=None,
                 precision=PARMES_REAL_GPU):
        super(GPUScalarField, self).__init__(self, queue, precision,
                                             parent, topology,
                                             name, idFromParent)
        self.data = np.asarray(self.data,
                               dtype=precision, order=ORDER)
        self.gpu_data = cl.Buffer(self.queue.context,
                                  cl.mem_flags.READ_WRITE,
                                  size=self.data.nbytes)
        self.mem_size = self.gpu_data.size
        if __VERBOSE__:
            print self.name, self.mem_size, "Bytes (",
            self.mem_size / (1024 ** 2), "MB)"
        else:
            print '.',

    @classmethod
    def fromScalarField(cls, queue, sfield, precision=PARMES_REAL_GPU):
        sfield.__class__ = cls
        GPUDiscreteField.__init__(sfield, queue, precision)
        sfield.data = np.asarray(sfield.data,
                                 dtype=precision, order=ORDER)
        sfield.gpu_data = cl.Buffer(sfield.queue.context,
                                    cl.mem_flags.READ_WRITE,
                                    size=sfield.data.nbytes)
        sfield.mem_size = sfield.gpu_data.size
        if __VERBOSE__:
            print sfield.name, sfield.mem_size, "Bytes (",
            print sfield.mem_size / (1024 ** 2), "MB)"
        else:
            print '.',

    @debug
    def initialize(self, formula=None):
        if self.initialization_kernel is None:
            ScalarField.initialize(self, formula)
            self.data = np.asarray(self.data,
                                   dtype=self.precision, order=ORDER)
            data, time = self.toDevice()
            print "Transfers : ", self.mem_size, "Bytes transfered at ",
            print "{0:.3f} GBytes/sec".format(
                (self.mem_size * 1e-9) / time)
        else:
            coord_min = np.ones(4, dtype=self.precision)
            mesh_size = np.ones(4, dtype=self.precision)
            coord_min[0:self.dimension] = np.asarray(self.topology.mesh.origin,
                                                     dtype=self.precision)
            mesh_size[0:self.dimension] = np.asarray(self.topology.mesh.space_step,
                                                     dtype=self.precision)
            evt = self.initialization_kernel(self.gpu_data,
                                             coord_min,
                                             mesh_size)
            self.queue.finish()
            cTime = (evt.profile.end - evt.profile.start) * 1e-9
            print "Computing  : ", cTime, "sec"
        self.contains_data = False
        self.data_on_device = True

    def toDevice(self):
        """
        Host to device method.

        @param queue : OpenCL queue.

        Performs a direct OpenCL copy from numpy arrays
        to OpenCL Buffers.\n
        Arrange memory on device so that vector components are
        contiguous in the direction of the component.\n
        Example : A 3D vector field F(x,y,z) is made up of 3
        OpenCL Buffers Fx, Fy, Fz. The memory layout is :
        - Fx : x-major ordering. On device,
        Fx[i + j*WIDTH + k*WIDTH*HEIGHT] access to Fx(i,j,k)
        - Fy : y-major ordering. On device,
        Fy[i + j*WIDTH + k*WIDTH*HEIGHT] access to Fy(j,i,k)
        - Fz : z-major ordering. On device,
        Fz[i + j*WIDTH + k*WIDTH*HEIGHT] access to Fz(k,i,j)
        """
        print "host->device :", self.name,
        time = 0.
        evt = [None, None, None]
        evt[0] = cl.enqueue_copy(self.queue, self.gpu_data,
                                 self.data.ravel(order=ORDER))
        self.queue.finish()
        for e in evt:
            if e is not None:
                time += (e.profile.end - e.profile.start) * 1e-9
        if __VERBOSE__:
            print self.mem_size, "Bytes transfered at ",
            print "{0:.3f} GBytes/sec".format(
                self.mem_size / (time * 1024 ** 3))
        else:
            print 'Transfer host->device'
        return self.mem_size, time

    def toHost(self):
        """
        Device to host method.

        @param queue : OpenCL queue.
        @param discreteField : Variable to transfer.

        Performs a direct OpenCL copy from OpenCL Buffers
        to numpy arrays.\n
        As memory layout is arranged on device, not only a
        copy is performed but also transpositions to have numpy
        arrays consistent to each other.
        """
        if not self.contains_data:
            print "device->host :", self.name,
            time = 0.
            evt = [None, None, None]
            evt[0] = cl.enqueue_copy(self.queue, self.data,
                                     self.gpu_data)
            self.queue.finish()
            for e in evt:
                if e is not None:
                    time += (e.profile.end - e.profile.start) * 1e-9
            if __VERBOSE__:
                print self.mem_size, "Bytes transfered at ",
                print "{0:.3f} GBytes/sec".format(
                    self.mem_size / (time * 1024 ** 3))
            else:
                print 'Transfer device->host'
            return self.mem_size, time
