"""
@package parmepy.fields.scalar

Discrete variable scalar description.
"""
from parmepy import __VERBOSE__
from parmepy.constants import np, ORDER, PARMES_REAL, debug
from parmepy.fields.discrete import DiscreteField


class ScalarField(DiscreteField):
    """
    A scalar field, defined on each grid of each
    subdomain of the given topology.

    """

    @debug
    def __init__(self, parent, topology=None, name="?", idFromParent=None):
        """Constructor for a scalar field"""
        super(ScalarField, self).__init__(parent, topology,
                                          name, idFromParent)
        self.vector = False
        self.data = np.zeros((self.resolution), dtype=PARMES_REAL, order=ORDER)

    def __getitem__(self, i):
        """
        Access to the content of the field.
        @param i : requested index.
        @return data[i].
        """
        return self.data.__getitem__(i)

    def __setitem__(self, i, value):
        """
        Set the content of the field component at position i.

        @param i : requested index.
        @param value : value to set.

        Usage :\n
        A = ScalarField(topo)\n
        A[2,1,1] = 12.0 # Calls A.data[2,1,1] = 12.0
        """
        self.data.__setitem__(i, value)

    @debug
    def initialize(self, formula=None):
        """
        Initialize values with given formula.

        @param formula : formula to apply on coordinates.

        Formula need to be vectorized by numpy.
        Therefore formula must have the following signature:
        @li 3D : float formula(float x, float y, float z)
        @li 2D : float formula(float x, float y)
        where x,y,z are point coordinates.
        """
        if formula is not None:
            if __VERBOSE__:
                print "...",
            v_formula = np.vectorize(formula)
            if self.dimension == 3:
                self.data = v_formula(*self.topology.mesh.coords)
            elif self.dimension == 2:
                self.data = v_formula(*self.topology.mesh.coords)
            else:
                self.data = v_formula(self.topology.mesh.coords)
            self.data = np.asarray(self.data,
                                   dtype=PARMES_REAL, order=ORDER)
            self.contains_data = True
        else:
            if __VERBOSE__:
                print "No formula",

    def __str__(self):
        """ Display class information. """
        s = '[' + str(self.topology.rank) + '] '
        s += self.name + " " + str(self.dimension)
        s += 'D discrete field with resolution '
        s += str(self.data.shape) + "."
        return s + "\n"


if __name__ == "__main__":
    print __doc__
    print "- Provided class : Scalar"
    print ScalarField.__doc__
