"""
Testing parmepy.field.scalar.ScalarField,
 parmepy.field.scalar.VectorField
"""
import parmepy as pp
from parmepy.constants import np
from parmepy.fields.continuous import Field
from parmepy.fields.analytical import AnalyticalField


def test_discretization():
    """Test multiple discretizations"""
    dom = pp.box.Box()
    csf = Field(dom)
    cvf = Field(dom, vector=True)
    resolTopo = [33, 33, 17]
    topo = pp.topology.Cartesian(dom, 3, resolTopo)
    csf.discretize(topo)
    cvf.discretize(topo)
    assert np.equal(csf.discreteField[0].resolution,
                    csf.discreteField[0].data.shape).all()
    assert not csf.discreteField[0].contains_data
    assert np.equal(cvf.discreteField[0].resolution,
                    cvf.discreteField[0].data[0].shape).all()
    assert np.equal(cvf.discreteField[0].resolution,
                    cvf.discreteField[0].data[1].shape).all()
    assert np.equal(cvf.discreteField[0].resolution,
                    cvf.discreteField[0].data[2].shape).all()
    assert not cvf.discreteField[0].contains_data


def test_init_scalar():
    dom = pp.box.Box()
    csf = Field(dom)
    resolTopo = [33, 11, 17]
    topo = pp.topology.Cartesian(dom, 3, resolTopo)
    csf.discretize(topo)
    scalar = lambda x, y, z: x * y * z
    true_data = np.zeros_like(csf.discreteField[0].data)
    for i in np.arange(resolTopo[0] - 1):
        for j in np.arange(resolTopo[1] - 1):
            for k in np.arange(resolTopo[2] - 1):
                true_data[i, j, k] = scalar(i / ((resolTopo[0] - 1) * 1.),
                                            j / ((resolTopo[1] - 1) * 1.),
                                            k / ((resolTopo[2] - 1) * 1.))
    csf.discreteField[0].initialize(scalar)
    assert np.allclose(csf.discreteField[0].data, true_data)


def test_init_vector():
    dom = pp.box.Box()
    cvf = Field(dom, vector=True)
    resolTopo = [33, 11, 17]
    topo = pp.topology.Cartesian(dom, 3, resolTopo)
    cvf.discretize(topo)
    vector = lambda x, y, z: (x ** 2, 1.0 - y, 10 * z + 10.)
    true_data_x = np.zeros_like(cvf.discreteField[0].data[0])
    true_data_y = np.zeros_like(cvf.discreteField[0].data[1])
    true_data_z = np.zeros_like(cvf.discreteField[0].data[2])
    for i in np.arange(resolTopo[0] - 1):
        for j in np.arange(resolTopo[1] - 1):
            for k in np.arange(resolTopo[2] - 1):
                true_data_x[i, j, k], \
                    true_data_y[i, j, k], \
                    true_data_z[i, j, k] = \
                    vector(i / ((resolTopo[0] - 1) * 1.),
                           j / ((resolTopo[1] - 1) * 1.),
                           k / ((resolTopo[2] - 1) * 1.))
    cvf.discreteField[0].initialize(vector)
    assert np.allclose(cvf.discreteField[0].data[0], true_data_x)
    assert np.allclose(cvf.discreteField[0].data[1], true_data_y)
    assert np.allclose(cvf.discreteField[0].data[2], true_data_z)


def test_accessing_vector_element():
    """Vector element access"""
    dom = pp.box.Box()
    cvf = Field(dom, vector=True)
    resolTopo = [33, 11, 17]
    topo = pp.topology.Cartesian(dom, 3, resolTopo)
    cvf.discretize(topo)
    vector = lambda x, y, z: (x ** 2, 1.0 - y, 10 * z + 10.)
    true_data_x = np.zeros_like(cvf.discreteField[0].data[0])
    true_data_y = np.zeros_like(cvf.discreteField[0].data[1])
    true_data_z = np.zeros_like(cvf.discreteField[0].data[2])
    cvf.discreteField[0].initialize(vector)
    print cvf.discreteField[0][:, :, :].shape
    tshape = (3, resolTopo[0] - 1, resolTopo[1] - 1, resolTopo[2] - 1)
    assert np.equal(cvf.discreteField[0][:, :, :].shape, tshape).all()
    tshape = (resolTopo[0] - 1, resolTopo[1] - 1, resolTopo[2] - 1)
    assert np.equal(cvf.discreteField[0][0][:, :, :].shape, tshape).all()
    tshape = (3,)
    assert np.equal(cvf.discreteField[0][0, 0, 0].shape, tshape).all()
    for i in np.arange(resolTopo[0] - 1):
        for j in np.arange(resolTopo[1] - 1):
            for k in np.arange(resolTopo[2] - 1):
                true_data_x[i, j, k], \
                    true_data_y[i, j, k], \
                    true_data_z[i, j, k] = \
                    vector(i / ((resolTopo[0] - 1) * 1.),
                           j / ((resolTopo[1] - 1) * 1.),
                           k / ((resolTopo[2] - 1) * 1.))
                assert np.allclose(cvf.discreteField[0][0][i, j, k],
                                   true_data_x[i, j, k])
                assert np.allclose(cvf.discreteField[0][1, i, j, k],
                                   true_data_y[i, j, k])
                assert np.allclose(cvf.discreteField[0][i, j, k],
                                   [true_data_x[i, j, k],
                                    true_data_y[i, j, k],
                                    true_data_z[i, j, k]])
    assert np.allclose(cvf.discreteField[0][0], true_data_x)
    assert np.allclose(cvf.discreteField[0][1], true_data_y)
    assert np.allclose(cvf.discreteField[0][2], true_data_z)
    assert np.allclose(cvf.discreteField[0][0][:, 1, 2], true_data_x[:, 1, 2])
    assert np.allclose(cvf.discreteField[0][0, :, 1, 2], true_data_x[:, 1, 2])
    #set item
    cvf.discreteField[0][1, 2, 3] = [1., 2., 3.]
    assert np.allclose(cvf.discreteField[0][1, 2, 3], [1., 2., 3.])
    cvf.discreteField[0][0, 1, :] = 5.
    true_data = [np.ones_like(cvf.discreteField[0].data[0][0, 1, :]) * 5.,
                 np.ones_like(cvf.discreteField[0].data[0][0, 1, :]) * 5.,
                 np.ones_like(cvf.discreteField[0].data[0][0, 1, :]) * 5.]
    assert np.allclose(cvf.discreteField[0][0, 1, :], true_data)
    cvf.discreteField[0][0][..., 9] = -1.
    assert np.allclose(cvf.discreteField[0][0][..., 9],
                       np.ones_like(true_data_x[..., 9]) * -1.)


