/**
 * 2D advection kernel.
 *
 * Memory reads and writes are performed using 4 element OpenCL vectors (float4 or double4).
 * Computations are done using builtin functions on vector types.
 * A local array is used un local memory as a cache for velocity.
 *
 * @param gvelo Velocity.
 * @param ppos Particle position.
 * @param dt Time step.
 * @param min_position Domain lower point.
 * @param dx Space step.
 */
__kernel void advection(__global const float* gvelo,
			__global float* ppos,
			float dt, float min_position, float dx)
{
  uint gidX = get_global_id(0);
  uint gidY = get_global_id(1);
  int4 ind, i_ind, i_ind_p;
  float invdx = 1.0/dx;
  uint i;
  float4 v, vp, y, p, coord;
  __local float velocity_cache[WIDTH];

  for(i=gidX*4; i<WIDTH; i+=WGNADVEC*4)
    {
      v = vload4((i+gidY*WIDTH)/4, gvelo);
      velocity_cache[i] = v.x;
      velocity_cache[i+1] = v.y;
      velocity_cache[i+2] = v.z;
      velocity_cache[i+3] = v.w;
    }

  barrier(CLK_LOCAL_MEM_FENCE);

  for(i=gidX*4; i<WIDTH; i+=WGNADVEC*4)
    {
      v = vload4(i/4, velocity_cache);
      coord = (float4)(i*dx, (i+1)*dx, (i+2)*dx, (i+3)*dx);
      p = fma((float4)(0.5*dt),v,coord);
      ind = convert_int4_rtn(p * invdx);
      y = (fma(- convert_float4(ind),dx,p))* invdx ;
      i_ind = ((ind + WIDTH) % WIDTH);
      i_ind_p = ((i_ind + 1) % WIDTH);
      v = (float4)(velocity_cache[i_ind.x], velocity_cache[i_ind.y],
                   velocity_cache[i_ind.z], velocity_cache[i_ind.w]);
      vp = (float4)(velocity_cache[i_ind_p.x], velocity_cache[i_ind_p.y],
                    velocity_cache[i_ind_p.z], velocity_cache[i_ind_p.w]);
      p=fma(mix(v,vp,y),dt,coord + (float4)(min_position));
      vstore4(p, (i+gidY*WIDTH)/4, ppos);
    }
}
