/**
 * Local memory as a 2D array to avoid bank conflicts.
 *
 * @param id Index 1D
 * @param nb_part Particle number per work-item
 *
 * @return Index in a 2D space.
 */
inline uint noBC_id(int id, int nb_part){
  return (id%nb_part)*WGNREMESH+(id/nb_part);
}

/**
  * 2D Remeshing kernel.
  *
  * @param ppos Particle position
  * @param pscal Particle scalar
  * @param gscal Grid scalar
  * @param min_position Domain lower point
  * @param dx Space step
  */
__kernel void remeshing(__global const float* ppos,
			__global const float* pscal,
			__global float* gscal, float min_position, float dx)
{
  uint gidX = get_global_id(0);
  uint gidY = get_global_id(1);
  float invdx = 1.0/dx;
  int4 ind;
  uint i, nb_part = WIDTH/WGNREMESH;
  float4 p, s, y;
  uint4 index4;

  __local float gscal_loc[WIDTH];

  for(i=gidX*4; i<WIDTH; i+=(WGNREMESH*4))
    {
      gscal_loc[i] = 0.0;
      gscal_loc[i+1] = 0.0;
      gscal_loc[i+2] = 0.0;
      gscal_loc[i+3] = 0.0;
    }
  barrier(CLK_LOCAL_MEM_FENCE);
  for(i=gidX*nb_part; i<(gidX + 1)*nb_part; i+=4)
    {
      p = vload4((i + gidY*WIDTH)/4, ppos) - (float4)(min_position);
      s = vload4((i + gidY*WIDTH)/4, pscal);
      ind = convert_int4_rtn(p * invdx);
      y = (p - convert_float4(ind) * dx) * invdx;
      index4 = convert_uint4((ind - 2 + WIDTH) % WIDTH);
      gscal_loc[noBC_id(index4.x,nb_part)] += (alpha(y.x,s.x));
      gscal_loc[noBC_id(index4.y,nb_part)] += (alpha(y.y,s.y));
      gscal_loc[noBC_id(index4.z,nb_part)] += (alpha(y.z,s.z));
      gscal_loc[noBC_id(index4.w,nb_part)] += (alpha(y.w,s.w));
      barrier(CLK_LOCAL_MEM_FENCE);
      index4 = (index4 + 1) % WIDTH;
      gscal_loc[noBC_id(index4.x,nb_part)] += (beta(y.x,s.x));
      gscal_loc[noBC_id(index4.y,nb_part)] += (beta(y.y,s.y));
      gscal_loc[noBC_id(index4.z,nb_part)] += (beta(y.z,s.z));
      gscal_loc[noBC_id(index4.w,nb_part)] += (beta(y.w,s.w));
      barrier(CLK_LOCAL_MEM_FENCE);
      index4 = (index4 + 1) % WIDTH;
      gscal_loc[noBC_id(index4.x,nb_part)] += (gamma(y.x, s.x));
      gscal_loc[noBC_id(index4.y,nb_part)] += (gamma(y.y, s.y));
      gscal_loc[noBC_id(index4.z,nb_part)] += (gamma(y.z, s.z));
      gscal_loc[noBC_id(index4.w,nb_part)] += (gamma(y.w, s.w));
      barrier(CLK_LOCAL_MEM_FENCE);
      index4 = (index4 + 1) % WIDTH;
      gscal_loc[noBC_id(index4.x,nb_part)] += (delta(y.x, s.x));
      gscal_loc[noBC_id(index4.y,nb_part)] += (delta(y.y, s.y));
      gscal_loc[noBC_id(index4.z,nb_part)] += (delta(y.z, s.z));
      gscal_loc[noBC_id(index4.w,nb_part)] += (delta(y.w, s.w));
      barrier(CLK_LOCAL_MEM_FENCE);
      index4 = (index4 + 1) % WIDTH;
      gscal_loc[noBC_id(index4.x,nb_part)] += (eta(y.x, s.x));
      gscal_loc[noBC_id(index4.y,nb_part)] += (eta(y.y, s.y));
      gscal_loc[noBC_id(index4.z,nb_part)] += (eta(y.z, s.z));
      gscal_loc[noBC_id(index4.w,nb_part)] += (eta(y.w, s.w));
      barrier(CLK_LOCAL_MEM_FENCE);
      index4 = (index4 + 1) % WIDTH;
      gscal_loc[noBC_id(index4.x,nb_part)] += (zeta(y.x, s.x));
      gscal_loc[noBC_id(index4.y,nb_part)] += (zeta(y.y, s.y));
      gscal_loc[noBC_id(index4.z,nb_part)] += (zeta(y.z, s.z));
      gscal_loc[noBC_id(index4.w,nb_part)] += (zeta(y.w, s.w));
      barrier(CLK_LOCAL_MEM_FENCE);
    }
  barrier(CLK_LOCAL_MEM_FENCE);
  for(i=gidX*4; i<WIDTH; i+=(WGNREMESH*4))
    vstore4((float4)(gscal_loc[noBC_id(i,nb_part)],gscal_loc[noBC_id(i+1,nb_part)], gscal_loc[noBC_id(i+2,nb_part)], gscal_loc[noBC_id(i+3,nb_part)]), (i + gidY*WIDTH)/4, gscal);
}
