"""
@package parmepy.gpu.tools

Tools for gpu management.
"""
from parmepy import __VERBOSE__
from parmepy.gpu import cl, PARMES_REAL_GPU

__platform, __device, __ctx, __queue = None, None, None, None
__precision = None


def get_opencl_environment(platform_id, device_id, device_type, precision):
    global __platform
    global __device
    global __ctx
    global __queue
    global __precision
    if __precision is None and not precision is None:
        __precision = precision
    if __queue is None:
        if __VERBOSE__:
            print "=== OpenCL environment ==="
        #Get platform.
        try:
            ## OpenCL platform
            __platform = cl.get_platforms()[platform_id]
        except IndexError:
            print "  Incorrect platform_id :", platform_id, ".",
            print " Only ", len(cl.get_platforms()), " available.",
            print " Getting defalut platform. "
            __platform = cl.get_platforms()[0]
        if __VERBOSE__:
            print "  Platform   "
            print "  - Name       :", __platform.name
            print "  - Version    :", __platform.version
        #Get device.
        try:
            ## OpenCL device
            __device = __platform.get_devices(
                eval("cl.device_type." + device_type.upper()))[device_id]
        except cl.RuntimeError as e:
            print "RuntimeError:", e
            __device = cl.create_some_context().devices[0]
        except AttributeError as e:
            print "AttributeError:", e
            __device = cl.create_some_context().devices[0]
        if __VERBOSE__:
            print "  Device"
            print "  - Name                :",
            print __device.name
            print "  - Type                :",
            print cl.device_type.to_string(__device.type)
            print "  - C Version           :",
            print __device.opencl_c_version
            print "  - Global mem size     :",
            print __device.global_mem_size / (1024 ** 3), "GB"
            print "===\n"
        #Creates GPU Context
        ## OpenCL context
        __ctx = cl.Context([__device])
        #Create CommandQueue on the GPU Context
        ## OpenCL command queue
        __queue = cl.CommandQueue(
            __ctx, properties=cl.command_queue_properties.PROFILING_ENABLE)
    return __platform, __device, __ctx, __queue


def _get_precision():
    global __precision
    if __precision is None:
        __precision = PARMES_REAL_GPU
    return __precision


def build_src(device, ctx, src, options, precision):
    if __VERBOSE__:
        print "=== Kernel sources compiling ==="
    if precision is PARMES_REAL_GPU:
        prg = cl.Program(ctx, src.replace('.0', '.0f'))
    else:
        prg = cl.Program(ctx, src.replace('float', 'double'))
    ## OpenCL program
    build = prg.build(options)
    if __VERBOSE__:
        print options
        print "Compiler options : ",
        print build.get_build_info(
            device, cl.program_build_info.OPTIONS)
        print "Compiler status : ",
        print build.get_build_info(
            device, cl.program_build_info.STATUS)
        print "Compiler log : ",
        print build.get_build_info(device,
                                   cl.program_build_info.LOG)
        print "===\n"
    return build


def _set_WorkItems(resolution, vector_width=1):
    ## Optimal work item number
    if len(resolution) == 3:
        workItemNumber = 64 if min(resolution) >= 64 \
            else min(resolution)
    else:
        workItemNumber = 256 if min(resolution) >= 256 \
            else min(resolution)
    ## Change work-item regarding problem size
    if resolution[0] % workItemNumber > 0:
        if len(resolution) == 3:
            print "Warning : GPU best performances obtained for",
            print "problem sizes multiples of 64"
        else:
            print "Warning : GPU best performances obtained for",
            print "problem sizes multiples of 256"
    while(resolution[0] % workItemNumber > 0):
        workItemNumber = workItemNumber / 2
    ## Change work-item regarding vector_width
    if workItemNumber * vector_width > resolution[0]:
        if resolution[0] % vector_width > 0:
            raise ValueError(
                "Resolution ({0}) must be a multiple of {1}".format(
                    resolution[0], vector_width))
        workItemNumber = resolution[0] // vector_width
    if len(resolution) == 3:
        gwi = (int(workItemNumber),
               int(resolution[1]), int(resolution[2]))
        lwi = (int(workItemNumber), 1, 1)
    else:
        gwi = (int(workItemNumber),
               int(resolution[1]))
        lwi = (int(workItemNumber), 1)
    return workItemNumber, gwi, lwi


def explore():
    """Explore hardware."""
    print "OpenCL exploration : "
    platforms = cl.get_platforms()
    platforms_info = ["name", "version", "vendor", "profile", "extensions"]
    devices_info = ["name", "version", "vendor", "profile", "extensions",
                    "available", "type", "compiler_available", "double_fp_config", "single_fp_config",
                    "global_mem_size", "global_mem_cache_type", "global_mem_cache_size", "global_mem_cacheline_size",
                    "local_mem_size", "local_mem_type", "max_clock_frequency", "max_compute_units",
                    "max_constant_buffer_size", "max_mem_alloc_size", "max_work_group_size",
                    "max_work_item_dimensions", "max_work_item_sizes",
                    "preferred_vector_width_double", "preferred_vector_width_float", "preferred_vector_width_int"]
    for pltfm in platforms:
        print "Platform:", pltfm.name
        for pltfm_info in platforms_info:
            print "  |-", pltfm_info, ':', eval("pltfm." + pltfm_info)
        devices = pltfm.get_devices()
        for dvc in devices:
            print "  |- Device:", dvc.name
            for dvc_info in devices_info:
                print "    |-", dvc_info, ':', eval("dvc." + dvc_info)


def _check_double_precision_capability(device, prec):
    ## Precision supported
    if __VERBOSE__:
        print " Precision capability  ",
    capable = True
    if prec is PARMES_REAL_GPU:
        if __VERBOSE__:
            print "for Single Precision: "
        for v in ['DENORM', 'INF_NAN',
                  'ROUND_TO_NEAREST', 'ROUND_TO_ZERO', 'ROUND_TO_INF',
                  'FMA', 'CORRECTLY_ROUNDED_DIVIDE_SQRT', 'SOFT_FLOAT']:
            try:
                if eval('(device.single_fp_config &' +
                        ' cl.device_fp_config.' +
                        v + ') == cl.device_fp_config.' + v):
                    if __VERBOSE__:
                        print v
            except AttributeError as ae:
                if v is 'CORRECTLY_ROUNDED_DIVIDE_SQRT':
                    capable = False
                if __VERBOSE__:
                    print v, 'is not supported in OpenCL C 1.2.\n',
                    print '   Exception catched : ', ae
    else:
        if __VERBOSE__:
            print "for Double Precision: "
        if device.double_fp_config > 0:
            for v in ['DENORM', 'INF_NAN', 'FMA',
                      'ROUND_TO_NEAREST', 'ROUND_TO_ZERO', 'ROUND_TO_INF',
                      'CORRECTLY_ROUNDED_DIVIDE_SQRT', 'SOFT_FLOAT']:
                try:
                    if eval('(device.double_fp_config &' +
                            ' cl.device_fp_config.' +
                            v + ') == cl.device_fp_config.' + v):
                        if __VERBOSE__:
                            print v
                except AttributeError as ae:
                    if v is 'CORRECTLY_ROUNDED_DIVIDE_SQRT':
                        capable = False
                    if __VERBOSE__:
                        print  v, 'is supported in OpenCL C 1.2.\n',
                        print '   Exception catched : ', ae
        else:
            raise ValueError("Double Precision is not supported by device")
    return capable
