"""
@package parmepy.mpi.mesh
Cartesian mesh class for local mesh.
@todo Remove LocalMesh or SubMesh
@todo rename to CartesianMesh
@todo write tests in parmepy/domain/tests/test_mesh.py
"""
from parmepy.constants import np, PARMES_REAL, PARMES_INTEGER, \
    debug


class LocalMesh(object):
    """
    Compute a local mesh resolution, using MPI process grid (topology)
    and the global mesh resolution.
    It also depends on ghost layer witdh and boundary conditions type.
    """
    def __init__(self, rank, resolution, start,
                 dom_size, dom_origin, ghosts=np.zeros((3))):
        # Current process rank in the topology that owns this mesh
        self.rank = rank
        # Local resolution
        self.resolution = resolution.copy() + 2 * ghosts
        # index of the lower point (in each dir) in the global mesh
        self.start = start  # - ghosts
        # index of the upper point (in each dir), global mesh
        self.end = self.start + self.resolution - 1
        # Mesh step size
        self.size = dom_size
        # Mesh coordinates
        coord_start = self.start * self.size + dom_origin - ghosts * self.size
        self.origin = coord_start
        coord_end = (self.end + 1) * self.size \
            + dom_origin - ghosts * self.size
#        print 'Print TOPO'
#        print 'resolution', self.resolution , 'shape',
#        (coord_end-coord_start)/self.size, 'coord_end',
# coord_end, 'coord_start',coord_start
        if self.start.size == 3:
            self.coords = np.ogrid[coord_start[0]:coord_end[0]:self.size[0],
                                   coord_start[1]:coord_end[1]:self.size[1],
                                   coord_start[2]:coord_end[2]:self.size[2]]
        elif self.start.size == 2:
            self.coords = np.ogrid[coord_start[0]:coord_end[0]:self.size[0],
                                   coord_start[1]:coord_end[1]:self.size[1]]
        else:
            self.coords = np.ogrid[coord_start[0]:coord_end[0]:self.size[0]]

    def __str__(self):
        """ Local mesh display """
        s = '[' + str(self.rank) + '] Local mesh resolution :' \
            + str(self.resolution) + '\n'

        s += 'Global indices :' + str(self.start)
        s += ' -- ' + str(self.end) + '\n'
        return s


class SubMesh(object):
    """
    Description of a local mesh, on one process of
    the topology.
    """

    @debug
    def __new__(cls, *args, **kw):
        return object.__new__(cls, *args, **kw)

    @debug
    def __init__(self, topo, g_start, resolution):

        ## Topology that creates (and owns) this mesh
        self._topology = topo
        ## Local resolution of this mesh, including ghost points
        self.resolution = resolution
        ## Dimension of the mesh
        self.dim = self.resolution.size
        ## index of the lowest point of this mesh
        ## (in each dir) in the global mesh
        self.global_start = g_start
        ## index of the upper point (in each dir), global mesh
        self.global_end = self.global_start + self.resolution - 1
        ## Mesh step size
        self.space_step = topo.domain.length / \
            (topo.globalMeshResolution - 1)
        ## Mesh local indices, only for "computed" points
        ## (i.e. excluding ghosts)
        self.local_start = topo.ghosts.copy()
        self.local_end = self.resolution.copy() - topo.ghosts[:] - 1
        ## Coordinates of the "lowest" point of this mesh (including ghost)
        ## Warning FP : this strongly depends on where is the origin
        ## of the domain, of the type of boundary conditions
        ## and if origin is on ghosts or "real" points.
        self.origin = topo.domain.origin.copy()
        self.origin[:] += self.space_step[:] * \
            (self.global_start[:].astype(PARMES_INTEGER) - topo.ghosts[:])

        self.end = self.origin + self.space_step * (self.resolution - 1)
        if self.dim == 1:
            self.coords = np.linspace(self.origin, self.end, self.resolution)
        elif self.dim == 2:
            cx = np.linspace(self.origin[0], self.end[0],
                             self.resolution[0])[:, np.newaxis]
            cy = np.linspace(self.origin[1], self.end[1],
                             self.resolution[1])[np.newaxis, :]
            self.coords = tuple([cx, cy])
        elif self.dim == 3:
            cx = np.linspace(self.origin[0], self.end[0],
                             self.resolution[0])[:, np.newaxis, np.newaxis]
            cy = np.linspace(self.origin[1], self.end[1],
                             self.resolution[1])[np.newaxis, :, np.newaxis]
            cz = np.linspace(self.origin[2], self.end[2],
                             self.resolution[2])[np.newaxis, np.newaxis, :]
            self.coords = tuple([cx, cy, cz])

    def __str__(self):
        """ Sub mesh display """
        s = 'Coords:' + str(self._topology.coords[:])
        s += ' Sub mesh resolution:' + str(self.resolution) + '\n'
        s += 'Starting point global indices :' + str(self.global_start) + '\n'
        s += 'Local indices for "computed" points (excluding ghosts) :\n'
        s += '   start = ' + str(self.local_start)
        s += ', end = ' + str(self.local_end) + '\n'
        return s
