"""
@package parmepy.mpi.topology
MPI Topologies
@todo merge CartesianTopology and Topology
"""

from parmepy.constants import ORDER, np, PARMES_INTEGER, XDIR, YDIR, ZDIR, PARMES_INDEX, debug
from parmepy.mpi.mesh import LocalMesh, SubMesh
from main_var import main_comm, MPI, main_size
from itertools import count


class CartesianTopology(object):
    """
    Define an MPI cartesian topology with an associated
    mesh for each sub-domain.
    Keyword arguments :
    - domain : the geometry; it must be a box.
    - globalMeshResolution : global gridResolution
    - dim : dimension of the topology
    - idTopo : name of the topology (integer)
    - dims : topology layout
    - comm : MPI communicator used to create this topology
    - periods : periodicity of the topology (in each direction)
    - ghosts : number of points in the ghost layer

    WARNING FP : this should work for any dim between 1 and 3!

    """
    ## Counter of CartesianTopology instances to set idTopo
    _ids = count(0)

    def __init__(self, domain, resolution, dim=None, dims=None, comm=MPI.COMM_WORLD, periods=None, ghosts=None):
        ## Associated domain
        self.domain = domain
        # Compute the "optimal" processus distribution for each direction of the grid topology
        nbprocs = comm.Get_size()
        if(dims is None):
            assert(dim is not None)
            self.dims = np.asarray(MPI.Compute_dims(nbprocs, dim), dtype=PARMES_INDEX)
            # Topology dimension
            self.dim = dim
        else:
            self.dims = np.asarray(dims, dtype=PARMES_INDEX)
            self.dim = self.dims.size
        # Define the topology
        # default period status is periodic
        if(periods is None):
            self.periods = self.dim * (True,)
        else:
            self.periods = periods
        self.topo = comm.Create_cart(self.dims, periods=self.periods)

        # Size of the topology
        self.size = self.topo.Get_size()
        # Rank of the current process, in the new topology
        self.rank = self.topo.Get_rank()
        if (self.rank == 0) :
            print 'dims = ' , self.dims
        # Coordinates of the current process
        self.coords = self.topo.Get_coords(self.rank)
        # Neighbours
        #        self.neighbours = np.zeros((dim,2))
        self.down, self.up = self.topo.Shift(XDIR, 1)
        if(self.dim > 1):
            self.west, self.east = self.topo.Shift(YDIR, 1)
        if(self.dim > 2):
            self.south, self.north = self.topo.Shift(ZDIR, 1)
        if self.dim == 3:
            self.tabSort = np.array([[self.up, 0], [self.down, 1],
                                     [self.east, 2], [self.west, 3],
                                     [self.north, 4], [self.south, 5]])
        else:
            self.tabSort = np.array([[self.up, 0] , [self.down, 1],
                                     [self.east, 2], [self.west, 3]])
        tabSort1 = self.tabSort[self.tabSort[:,0]<=self.rank]
        tabSort2 = self.tabSort[self.tabSort[:,0]>self.rank]
        if (np.asarray(tabSort1[:,0].shape)>0):
            tabSort1 = tabSort1[tabSort1[:,0].argsort()]
        if (np.asarray(tabSort2[:,0].shape)>0):
            tabSort2 = tabSort2[tabSort2[:,0].argsort()]
            tabSort2 = tabSort2[np.invert(tabSort2[:,0].argsort() )]
        self.tabSort = np.concatenate((tabSort1, tabSort2))
        self.tabSort=self.tabSort[self.tabSort[:,0].argsort()]
        # ghost layer (default = 0)
        if(ghosts is None):
            self.ghosts = np.zeros((self.domain.dimension), dtype=PARMES_INDEX)
        else:
            self.ghosts=np.asarray(ghosts, dtype=PARMES_INDEX)
        # Global resolution
        self.resolution = np.asarray(resolution, dtype=PARMES_INDEX)
        nbpoints = self.resolution[:self.dim] // self.dims
        remainingPoints = self.resolution[:self.dim] % self.dims
        localResolution = self.resolution.copy()
        localResolution[:self.dim] = nbpoints
        for i in range(self.dim):
            if(self.coords[i] == self.dims[i] - 1):
                localResolution[i] = nbpoints[i] + remainingPoints[i]
        start = np.zeros((domain.dimension), dtype=PARMES_INDEX)
        start[:self.dim] = self.coords * nbpoints
        localResolution = localResolution - 1 * self.periods
        self.mesh = LocalMesh(self.rank,
                              resolution=localResolution,
                              start=start,
                              dom_size=self.domain.length / (self.resolution-1),
                              dom_origin=self.domain.origin,
                              ghosts=self.ghosts)

    def __str__(self):
        """ Topology info display """
        s = '=======================================================\n'
        if(self.rank == 0):
            s += str(self.dim) + 'D cartesian Topology of size ' + str(self.dims) + '\n'
        s += str(self.mesh)
        s += '=======================================================\n'
        return s


class Cartesian(object):
    """
    Define an MPI cartesian topology with an associated
    mesh for each sub-domain.
    Keyword arguments :
    - domain : the geometry; it must be a box.
    - globalMeshResolution : user input. Number of points in the domain
    in each direction. We assume that first point corresponds to origin, and
    last point to boundary point, whatever the boundary type is.
    That is x[0] = domain.origin and
    x[globalMeshResolution-1] = domain.Lengths_x.
    - dim : dimension of the topology
    - dims : topology resolution (i.e process layout in each direction).
    - comm : MPI communicator used to create this topology
    - periods : periodicity of the topology (in each direction)
    - ghosts : number of points in the ghost layer

    WARNING FP : this should work for any dim between 1 and 3!

    """

    @debug
    def __new__(cls, *args, **kw):
        return object.__new__(cls, *args, **kw)

    ## Counter of topology.Cartesian instances to set idTopo.
    __ids = count(0)

    @debug
    def __init__(self, domain, dim, globalMeshResolution,
                 comm=main_comm, periods=None, ghosts=None, cutdir=None):
        ## Associated domain
        self.domain = domain
        ## An id for the topology
        self.idTopo = self.__ids.next()
        ## Topology dimension
        self.dim = dim
        ## Communicator used to build the topology
        self.comm = comm

        ## First of all we need to set dims (if not already set)
        if not hasattr(self, 'dims'):
            # If dims is not provided, computation of  the "optimal" processus
            # distribution for each direction of the grid topology.
            self.dims = np.asarray(MPI.Compute_dims(main_size, dim),
                                   dtype=PARMES_INDEX)
            ## Reorder dims according to the data layout
            # if arrays are in "fortran" order (column major)
            # the last dir has priority for distribution.
            # For C-like (row major) arrays, first dir is the
            # first to be distributed, and so on.
            self.dims.sort()
            if(ORDER == 'C'):
                self.dims[:] = self.dims[::-1]

        # default period status is periodic
        if(periods is None):
            ## Topology BC type
            self.periods = self.dim * (True,)
        else:
            assert (periods.size == self.dim)
            self.periods = periods

        self.topo = comm.Create_cart(self.dims, periods=self.periods)

        ## Size of the topology
        self.size = self.topo.Get_size()
        ## Rank of the current process, in the new topology
        self.rank = self.topo.Get_rank()
        ## Coordinates of the current process
        self.coords = np.asarray(self.topo.Get_coords(self.rank),
                                 dtype=PARMES_INDEX)

        ## Neighbours
        self.neighbours = np.zeros((2, self.dim))
        for direction in range(self.dim):
            self.neighbours[:, direction] = self.topo.Shift(direction, 1)

        #self.down, self.up = self.topo.Shift(XDIR, 1)
        #self.neighbours[:, XDIR] = (self.down, self.up)
        #if(domain.dimension > 1):
        #    self.west, self.east = self.topo.Shift(YDIR, 1)
#            self.neighbours[:, YDIR] = self.west, self.east

##         if(domain.dimension > 2):
##             self.south, self.north = self.topo.Shift(ZDIR, 1)
##             self.neighbours[:, ZDIR] = self.south, self.north
##             self.tabSort = np.array([[self.up,0] , [self.down,1] , [self.east,2], [self.west,3],
##                                      [self.north,4],[self.south,5]])
##             tabSort1 = self.tabSort[self.tabSort[:,0]<=self.rank]
##             tabSort2 = self.tabSort[self.tabSort[:,0]>self.rank]
##             if (np.asarray(tabSort1[:,0].shape)>0):
##                 tabSort1 = tabSort1[tabSort1[:,0].argsort()]
##                 if (np.asarray(tabSort2[:,0].shape)>0):
##                     #            print 'tabSort2', self.rank, tabSort2 , tabSort2[:,0].shape
##                     tabSort2 = tabSort2[np.invert(tabSort2[:,0].argsort() )]
##                     tabSort2 = tabSort2[tabSort2[:,0].argsort()]
##                     self.tabSort = np.concatenate((tabSort1, tabSort2))
##                     self.tabSort=self.tabSort[self.tabSort[:,0].argsort()]

         # ghost layer (default = 0)
        if(ghosts is None):
            self.ghosts = np.zeros((self.domain.dimension),
                                   dtype=PARMES_INDEX)
        else:
            self.ghosts = np.asarray(ghosts, dtype=PARMES_INDEX)
            assert(self.ghosts.size == domain.dimension)

        # --- Computation of the local mesh resolution and indices ---
        ##  Global Mesh Resolution
        self.globalMeshResolution = np.asarray(globalMeshResolution,
                                               dtype=PARMES_INDEX)

        # set cutdir if required (i.e. not set by withResolution constructor).
        if cutdir is None:  # not hasattr(self, '__cutdir'):
            self.__cutdir = np.zeros((domain.dimension), dtype=bool)
            self.__cutdir[:self.dim] = True
            if ORDER == 'F':
                self.__cutdir[:] = self.__cutdir[::-1]
        else:
            self.__cutdir = np.asarray(cutdir, dtype=bool)

        # Number of "computed" points (i.e. excluding ghosts/boundaries).
        pts_noghost = np.zeros((self.dim), dtype=PARMES_INDEX)
        # Warning FP : we should test boundary conditions type here
        # If periodic, resol_calc = (globalMeshResolution - 1)
        # else, resol_calc =  (globalMeshResolution - 2)
        resolCalc = np.zeros((domain.dimension), dtype=PARMES_INDEX)
        resolCalc[:] = self.globalMeshResolution[:] - 1
        pts_noghost[:] = resolCalc[self.__cutdir.nonzero()] // self.dims[:]

        # If any, remaining points are added on the mesh of the last process.
        remainingPts = np.zeros(self.dim, dtype=PARMES_INDEX)
        remainingPts[:] = resolCalc[self.__cutdir.nonzero()] % self.dims[:]

        # Then the total number of points (size of arrays to be allocated)
        nbpoints = pts_noghost.copy()
        for i in range(self.dim):
            if(self.coords[i] == self.dims[i] - 1):
                nbpoints[i] += remainingPts[i]

        self.localGridResolution = resolCalc.copy()
        self.localGridResolution[self.__cutdir.nonzero()] = nbpoints[:]
        self.localGridResolution[:] += 2 * self.ghosts[:]

        ## Global indices for the local mesh points
        ## self.G_start[i] is the index of the first point in the i direction.
        ## It includes ghosts/boundary points.
        self.G_start = np.zeros((domain.dimension), dtype=PARMES_INDEX)
        self.G_start[self.__cutdir.nonzero()] = self.coords[:] * pts_noghost[:]

        ## Local indices.
        ## L_start[i] is the index of the first point which need to be
        ## computed, in the i-direction. It means that ghosts/boundary
        ## points are NOT included.
        ##L_start = self.ghosts.copy()
        ##L_end = self.localGridResolution.copy() - self.ghosts[:] - 1
        self.mesh = None

    @debug
    def setUp(self):
        """Topology set up.

        Create topology's Local Mesh
        """
        self.mesh = SubMesh(self, self.G_start, self.localGridResolution)

        #        print "aiaiaiai", self.rank, self.coords, "aaa", self.localGridResolution, self.G_start, L_start, L_end
##         print "local ...",  self.rank, " ", self.localGridResolution
##         self.mesh = LocalMesh(self.rank,
##                               resolution=self.localGridResolution,
##                               start=start,
##                               size=self.domain.length / (self.globalMeshResolution - 1),
##                               dom_origin=self.domain.origin,
##                               ghosts=self.ghosts)

    @classmethod
    def withResolution(cls, domain, topoResolution,
                       globalMeshResolution, cutdir=None,
                       ghosts=None, periods=None):

        cls.dims = np.asarray(topoResolution, dtype=PARMES_INDEX)
        cls.dim = cls.dims.size
        assert cls.dims.all(), "Topology error : you try to \
            create a topology with 0 resolution in one direction."

        # We set three ways to choose which direction(s) will be distributed :
        # 1 - If the dimension of the topology is equal to the domain's one,
        # then cutdir is useless and we keep the order
        # provided by topoResolution.
        # 2 - topoDim < domain.dimension and no cutdir given by user :
        # C-like arrays : first dir first distributed, fortran-like
        # arrays: last dir first distributed.
        # 3 - topoDim < domain.distributed and cutdir as input argument :
        # we use cutdir to choose which direction(s) will be distributed.
        if cls.dim < domain.dimension:
            cls.dims.sort()
            if(ORDER == 'C'):
                cls.dims[:] = cls.dims[::-1]

        return cls(domain, cls.dim, globalMeshResolution, cutdir=cutdir,
                   periods=periods, ghosts=ghosts)

    @classmethod
    def withResolutionFixed(cls, domain, topoResolution,
                            globalMeshResolution, cutdir=None,
                            ghosts=None, periods=None):

        cls.dims = np.asarray(topoResolution, dtype=PARMES_INDEX)
        cls.dim = cls.dims.size
        assert cls.dims.all(), "Topology error : you try to \
            create a topology with 0 resolution in one direction."

        return cls(domain, cls.dim, globalMeshResolution,
                   cutdir=cutdir, periods=periods, ghosts=ghosts)

    def __eq__(self, other):
        """
        Topology comparison method. Comparison based on:
        @li globalMeshResolution
        @li localMeshResolution
        @li dims
        @li domain
        @todo completer le test dans parmepy/mpi/tests/test_topology.py
        """
        return np.equal(self.globalMeshResolution,
                        other.globalMeshResolution).all() and \
            np.equal(self.localGridResolution, other.localGridResolution).all() and \
            np.equal(self.dims, other.dims).all() and \
            np.equal(self.ghosts, other.ghosts).all() and \
            self.domain == other.domain

    def __str__(self):
        """ Topology info display """
        s = '======== Topology id : ' + str(self.idTopo) + ' ========\n'
        if(self.rank == 0):
            s += str(self.dim) + 'D cartesian Topology of size ' \
                + str(self.dims) + '\n'
        s += 'Process of coordinates ' + str(self.coords[:]) + '\n'
        s += 'Neighbours coordinates : \n'
        s += str(self.neighbours) + '\n'
        s += 'Ghost layer :' + str(self.ghosts[:]) + '\n'
        s += 'Process rank : ' + str(self.rank) + '\n'
        s += str(self.mesh)
        s += '=================================\n'
        return s


if __name__ == "__main__":
    print "This module defines the following classes:"
    print "- CartesianTopology: ", CartesianTopology.__doc__
