"""
@package parmepy.numerics.gpu_kernel
"""
from parmepy.constants import debug
from parmepy import __VERBOSE__
from parmepy.numerics.method import NumMethod
from parmepy.gpu import cl


class KernelListLauncher:
    """
    OpenCL kernel list launcher.

    Manage launching of OpenCL kernels as a list.
    """
    @debug
    def __init__(self, kernel, queue, gsize, lsize=None):
        """
        Create a kernel list launcher.
        @param kernel : kernel list.
        @param queue : OpenCL command queue.
        @param gsize : OpenCL global size index.
        @param lsize : OpenCL local size index.
        """
        ## OpenCL Kernel list
        self.kernel = kernel
        #print [k.function_name for k in self.kernel]
        ## OpenCL command queue
        self.queue = queue
        ## OpenCL global size index.
        self.global_size = gsize
        ## OpenCL local size index.
        self.local_size = lsize
        self.call_number = [0 for k in self.kernel]

    @debug
    def __call__(self, d, *args):
        """
        Launch a kernel.

        @param d : kernel index in kernel list.
        @param args : kernel arguments.
        @return OpenCL Event

        OpenCL global size and local sizes are not given in
        args. Class member are used.
        """
        return KernelListLauncher.launch_sizes_in_args(
            self, d, self.global_size[d], self.local_size[d], *args)

    @debug
    def launch_sizes_in_args(self, d, *args):
        """
        Launch a kernel.

        @param d : kernel index in kernel list.
        @param args : kernel arguments.
        @return OpenCL Event.

        Opencl global and local sizes are given in args.
        """
        if __VERBOSE__:
            print self.kernel[d].function_name, args[0], args[1]
        self.call_number[d] += 1
        evt = self.kernel[d](self.queue, *args)
        return evt

    def finish(self):
        self.queue.finish()

    def function_name(self, d=None):
        """Prints OpenCL Kernels function names informations"""
        if d is not None:
            return self.kernel[d].get_info(cl.kernel_info.FUNCTION_NAME)
        else:
            return [k.get_info(cl.kernel_info.FUNCTION_NAME)
                    for k in self.kernel]


class KernelLauncher(KernelListLauncher):
    """
    OpenCL kernel launcher.

    Manage launching of one OpenCL kernel as a KernelListLauncher
    with a list of one kernel.
    """
    @debug
    def __init__(self, kernel, queue, gsize, lsize):
        """
        Create a KernelLauncher.

        @param kernel : kernel.
        @param queue : OpenCL command queue.
        @param gsize : OpenCL global size index.
        @param lsize : OpenCL local size index.

        Create a KernelListLauncher with a list of one kernel.
        """
        KernelListLauncher.__init__(self, [kernel], queue, [gsize], [lsize])

    @debug
    def launch_sizes_in_args(self, *args):
        """
        Launch the kernel.

        @param args : kernel arguments.
        @return OpenCL Event.

        Opencl global and local sizes are given in args.
        """
        return KernelListLauncher.launch_sizes_in_args(self, 0, *args)

    def finish(self):
        """Wrapping OpenCL queue.finish() method."""
        self.queue.finish()

    @debug
    def __call__(self, *args):
        """
        Launch the kernel.

        @param args : kernel arguments.
        @return OpenCL Event

        OpenCL global size and local sizes are not given in args.
        Class member are used.
        """
        return KernelListLauncher.__call__(self, 0, *args)

    def function_name(self):
        """Prints OpenCL Kernel function name informations"""
        res = KernelListLauncher.function_name(self, 0)
        return res
