"""
@package ode_solver.runge_kutta3

RK3 method interface.
"""
from parmepy.constants import np, PARMES_REAL, ORDER
from parmepy.numerics.integrators.odesolver import ODESolver


class RK3(ODESolver):
    """
    ODESolver implementation for solving an equation system with RK3 method.
    y'(t) = f(t,y)

    y(t_n+1)= y(t_n) + dt*y'[y(t_n)+dt/2*y'(y(t_n))].

    """
    def __init__(self, f=None, dim=3):
        """
        Constructor.

        @param f function f(t,y) : Right hand side of the equation to solve.
        @param dim : dimensions
        @param conditions : function to apply boundary conditions.
        """
        ODESolver.__init__(self, f)
        self.name = 'RK3'

    def __call__(self, f, t, dt, y):
        """
        Integration step for RK2 Method.
         up = f[t, Y(t,space)].

        @param f : right-hand side.
        @param t : current time.
        @param dt : time step.
        @param y : function Y(t,space)
        """
        # K1 = dt *f[t,y(t,x,y,z)]
        K1 = dt * np.asarray(f(t, y), dtype=PARMES_REAL, order=ORDER)
        # result = y(t,x,y,z)] + dt *f[t + dt/2 , y(t,x,y,z) + k1/2.]
        K2 = dt * np.asarray(f(t + dt / 3., y + K1[:, :, :, :] / 3.),
                             dtype=PARMES_REAL, order=ORDER)
        K3 = dt * np.asarray(f(t + dt * 2. / 3., y + K2[:, :, :, :] * 2. / 3.),
                             dtype=PARMES_REAL, order=ORDER)
        # result = y(t,x,y,z)] + 1/4 [ k1 + 3 k3]
        result = y + 1. / 4. * (K1 + 3. * K3)
        return result

    def __str__(self):
        """ToString method"""
        return "RK3 Method"


if __name__ == "__main__":
    print __doc__
    print "- Provided class : RK3"
    print RK3.__doc__
