"""
@package parmepy.numerics.remeshing
"""
from parmepy.constants import np, PARMES_REAL, ORDER
from parmepy.numerics.remeshing.remeshing import Remeshing


class M6Prime(Remeshing):
    """Remshing with M6prime function"""

    def __init__(self, dim):
        Remeshing.__init__(self, dim)
        self.name = 'M6Prime'
        self._alpha = lambda y, s: s * y * (
            y * (y * (y * (13. - 5. * y) - 9.) - 1.) + 2.) / 24.
        self._beta = lambda y, s: s * y * (
            y * (y * (y * (25. * y - 64.) + 39.) + 16.) - 16.) / 24.
        self._gamma = lambda y, s: s * (
            y * y * (y * (y * (126. - 50. * y) - 70.) - 30.) / 24. + 1.)
        self._delta = lambda y, s: s * y * (
            y * (y * (y * (50. * y - 124.) + 66.) + 16.) + 16.) / 24.
        self._zeta = lambda y, s: s * y * (
            y * (y * (y * (61. - 25. * y) - 33.) - 1.) - 2.) / 24.
        self._eta = lambda y, s: s * y * y * y * (
            y * (5. * y - 12.) + 7.) / 24.

    def __call__(self, ppos, pscal, d):

        """
        Remesh particles at position ppos with scalar
        pscal along direction d.

        @param ppos : particle position
        @param pscal : particle scalar
        @param d : splitting direction
        """
        size = ppos.shape
        row, col = np.indices(size)
        row_i = np.array(row)
        floor = np.floor(ppos * size[d])
        y = np.array(ppos * size[d] - floor, dtype=PARMES_REAL, order=ORDER)
        result = np.zeros(size, dtype=PARMES_REAL, order=ORDER)
        ind = floor.astype(np.int32) % size[d]
        # alpha
        row = (ind - 2) % size[d]
        for i in xrange(size[d]):
            sl = self.slice_i_along_d(i, d)
            result[row[sl], col[sl]] = result[row[sl], col[sl]] +\
                self._alpha(y[sl], pscal[sl])
        # beta
        row = (ind - 1) % size[d]
        for i in xrange(size[d]):
            sl = self.slice_i_along_d(i, d)
            result[row[sl], col[sl]] = result[row[sl], col[sl]] +\
                self._beta(y[sl], pscal[sl])
        # gamma
        row = (ind) % size[d]
        for i in xrange(size[d]):
            sl = self.slice_i_along_d(i, d)
            result[row[sl], col[sl]] = result[row[sl], col[sl]] +\
                self._gamma(y[sl], pscal[sl])
        # delta
        row = (ind + 1) % size[d]
        for i in xrange(size[d]):
            sl = self.slice_i_along_d(i, d)
            result[row[sl], col[sl]] = result[row[sl], col[sl]] +\
                self._delta(y[sl], pscal[sl])
        # zeta
        row = (ind + 2) % size[d]
        for i in xrange(size[d]):
            sl = self.slice_i_along_d(i, d)
            result[row[sl], col[sl]] = result[row[sl], col[sl]] +\
                self._zeta(y[sl], pscal[sl])
        # eta
        row = (ind + 3) % size[d]
        for i in xrange(size[d]):
            sl = self.slice_i_along_d(i, d)
            result[row[sl], col[sl]] = result[row[sl], col[sl]] +\
                self._eta(y[sl], pscal[sl])
        return result
