"""
@package parmepy.operator.advection

Advection operator representation.
"""
from parmepy.constants import debug, np, PARMES_INDEX
from parmepy.operator.continuous import Operator
from parmepy.mpi.topology import Cartesian


class Advection(Operator):
    """
    Advection operator representation.
    """

    @debug
    def __init__(self, velocity, advectedFields, resolutions=None, method='',
                 **other_config):
        """
        Create an Transport operator from given variables velocity and scalar.

        @param velocity : velocity variable.
        @param scalar : scalar variable.
        """
        v = [velocity]
        if isinstance(advectedFields, list):
            [v.append(f) for f in advectedFields]
        else:
            v.append(advectedFields)
        Operator.__init__(self, v)
        ## Transport velocity
        self.velocity = velocity
        ## Transported fields
        self.advectedFields = advectedFields
        self.resolutions = resolutions
        self.method = method
        self.config = other_config

    @debug
    def setUp(self):
        """
        Transport operator discretization method.
        Create an discrete Transport operator from given specifications.

        """
        Operator.setUp(self)
        if self.method.find('scales') >= 0:
            if not self.advectedFields.dimension == 3:
                    raise ValueError("Scales Advection not implemented in 2D.")
            ## Scales imports for topology creation
            from parmepy.operator.scales_advection import ScalesAdvection
            from parmepy.f2py import scales2py as scales
            from parmepy.mpi.main_var import main_size

            ## Extract order form self.method (default p_O2)
            order = 'p_O2'
            for o in ['p_O2', 'p_O4', 'p_M6']:
                if self.method.find(o) >= 0:
                    order = o
            ## Scales nbcells equals resolutions - 1
            ## resolutions is a number of points given by user
            nbcells = np.asarray(self.resolutions[self.advectedFields],
                                 dtype=PARMES_INDEX) - 1
            topodims = [1, 1, main_size]
            scalesres, scalesoffset, stab_coeff = \
                scales.init_advection_solver(nbcells,
                                             self.domain.length,
                                             topodims, order=order)
            ## Use same topodims as scales to create Cartesian topology
            ## in order to discretize our fields
            for v in self.variables:
                topoId, topo = self.domain.addTopology(
                    Cartesian.withResolution(
                        domain=self.domain, topoResolution=topodims,
                        globalMeshResolution=self.resolutions[v]))
                df, dfId = v.discretize(topo)
                self.discreteFieldId[v] = dfId
            self.discreteOperator = ScalesAdvection(self, method=self.method,
                                                    **self.config)
        else:
            #variables discretization
            for v in self.variables:
                topoId, topo = self.domain.addTopology(
                    Cartesian(domain=self.domain, dim=self.domain.dimension,
                              globalMeshResolution=self.resolutions[v]))
                df, dfId = v.discretize(topo)
                self.discreteFieldId[v] = dfId
            if self.discreteOperator is None:
                if self.method.find('gpu_2k') >= 0:
                    if self.advectedFields.vector:
                        raise ValueError("Not implemented yet. GPU advection" +
                                         " is only for scalar fields")
                    from parmepy.operator.gpu_particle_advection_2k \
                        import GPUParticleAdvection2k
                    self.discreteOperator = \
                        GPUParticleAdvection2k(self, method=self.method,
                                               **self.config)
                elif self.method.find('gpu_1k') >= 0:
                    if self.advectedFields.vector:
                        raise ValueError("Not implemented yet. GPU advection" +
                                         " is only for scalar fields")
                    from parmepy.operator.gpu_particle_advection_1k \
                        import GPUParticleAdvection1k
                    self.discreteOperator = \
                        GPUParticleAdvection1k(self, method=self.method,
                                               **self.config)
                else:
                    print "Using default advection operator"
                    if self.advectedFields.vector:
                        raise ValueError("Not implemented yet. default" +
                                         " advection is only for " +
                                         "scalar fields")
                    from parmepy.operator.particle_advection \
                        import ParticleAdvection
                    self.discreteOperator = \
                        ParticleAdvection(self, method=self.method,
                                          **self.config)
        self.discreteOperator.setUp()

    def __str__(self):
        """ToString method"""
        s = "Advection operator (ContinuousOperator)"
        if self.discreteOperator is not None:
            s += " with the following discretization:\n"
            s += str(self.discreteOperator)
        else:
            s += ". Not discretised"
        return s + "\n"

if __name__ == "__main__":
    print __doc__
    print "- Provided class : Advection"
    print Advection.__doc__
