"""
@package parmepy.operator.analytic

Analytic operator representation.
"""
from parmepy.constants import debug
from parmepy.operator.continuous import Operator
from parmepy.mpi.topology import Cartesian
from parmepy.operator.velocity_d import Analytic_D


class Analytic(Operator):
    """
    Analytic operator representation.

    This operator apply an analytic formula, given by user on fields.
    """

    @debug
    def __init__(self, fields, formula, resolutions=None, method='',
                 **other_config):
        """
        Create an operator for computing analyically a formula on field(s).
        Apply the same formula on each fields.

        @param fields : field
        """
        if not callable(formula):
            raise ValueError("Analytic formula must be a function")
        Operator.__init__(self, [fields])
        ## velocity variable
        self.resolutions = resolutions
        self.method = method
        self.config = other_config
        self.formula = formula

    @debug
    def setUp(self):
        """
        Set up of Analytic operator.
        """
        Operator.setUp(self)
        # Variables discretization
        for v in self.variables:
            topoId, topo = self.domain.addTopology(
                Cartesian(domain=self.domain, dim=self.domain.dimension,
                          globalMeshResolution=self.resolutions[v]))
            df, dfId = v.discretize(topo)
            self.discreteFieldId[v] = dfId
        if self.method.find('gpu') >= 0:
            from parmepy.operator.gpu_analytic import GPUAnalytic
            self.discreteOperator = GPUAnalytic(self, method=self.method,
                                                **self.config)
        else:
            self.discreteOperator = Analytic_D(self, method=self.method,
                                               **self.config)
        self.discreteOperator.setUp()

    def __str__(self):
        """ToString method"""
        s = "Velocity operator (ContinuousOperator)"
        if self.discreteOperator is not None:
            s += " with the following discretization:\n"
            s += str(self.discreteOperator)
        else:
            s += ". Not discretised"
        return s + "\n"

if __name__ == "__main__":
    print __doc__
    print "- Provided class : Advection"
    print Advection.__doc__
