# -*- coding: utf-8 -*-
"""
@package operator
Discrete Diffusion operator using FFT
"""
from parmepy.f2py import fftw2py
from parmepy.operator.discrete import DiscreteOperator
from parmepy.constants import np, ORDER, PARMES_REAL, debug
import time


class DiffusionFFT(DiscreteOperator):
    """
    Operator representation.
    """

    @debug
    def __init__(self, diff, method='',
                 viscosity=0.1, with_curl=False):
        """
        Constructor.
        @param velocity : descretized velocity to use.
        @param vorticity : discretized vorticity to update.
        """
        DiscreteOperator.__init__(self, diff)
        ## Parent Continous Op.
        self.diff = diff
        ## Velocity.
        self.velocity = self.diff.velocity.discreteField[
            self.diff.discreteFieldId[self.diff.velocity]]
        ## Vorticity.
        self.vorticity = self.diff.vorticity.discreteField[
            self.diff.discreteFieldId[self.diff.vorticity]]
        ## Viscosity.
        self.viscosity = viscosity
        self.domain = self.diff.domain
        ## Boolean : pure vort diffusion or curl(u) + vort diffusion.
        self.with_curl = with_curl
        self.compute_time = 0.

    @debug
    def setUp(self):
        pass

    @debug
    def apply(self, t, dt, ite):
        """
        Apply operator.
        """
        self.compute_time = time.time()

        if (self.domain.dimension == 2):

            raise ValueError("Not yet implemented")

        elif (self.domain.dimension == 3):

            if self.with_curl:
                self.vorticity.data[0], \
                self.vorticity.data[1], \
                self.vorticity.data[2] = \
                    fftw2py.solve_curl_diffusion_3d(self.viscosity * dt,
                                                    self.velocity.data[0],
                                                    self.velocity.data[1],
                                                    self.velocity.data[2],
                                                    self.vorticity.data[0],
                                                    self.vorticity.data[1],
                                                    self.vorticity.data[2])

            else:
                self.vorticity.data[0], \
                self.vorticity.data[1], \
                self.vorticity.data[2] = \
                    fftw2py.solve_diffusion_3d(self.viscosity * dt,
                                               self.vorticity.data[0],
                                               self.vorticity.data[1],
                                               self.vorticity.data[2])

        else:
            raise ValueError("invalid problem dimension")


#        ind0a = self.topology.mesh.local_start[0]
#        ind0b = self.topology.mesh.local_end[0] + 1
#        ind1a = self.topology.mesh.local_start[1]
#        ind1b = self.topology.mesh.local_end[1] + 1
#        ind2a = self.topology.mesh.local_start[2]
#        ind2b = self.topology.mesh.local_end[2] + 1

#        vorticityNoG = [np.zeros((self.resolution - 2 * self.ghosts),
#                                 dtype=PARMES_REAL,
#                                 order=ORDER) for d in xrange(self.dim)]
#        velocityNoG = [np.zeros((self.resolution - 2 * self.ghosts),
#                                dtype=PARMES_REAL,
#                                order=ORDER) for d in xrange(self.dim)]
#        for i in xrange(self.dim):
#            vorticityNoG[i][...] = self.vorticity[i][ind0a:ind0b,
#                                                     ind1a:ind1b, ind2a:ind2b]
#            velocityNoG[i][...] = self.velocity[i][ind0a:ind0b,
#                                                   ind1a:ind1b, ind2a:ind2b]

#        # Curl + Vorticity diffusion
##        vorticityNoG[0][...], vorticityNoG[1][...], vorticityNoG[2][...] = \
##            fftw2py.solve_curl_diffusion_3d(self.viscosity * dt,
##                                       velocityNoG[0][...],
##                                       velocityNoG[1][...],
##                                       velocityNoG[2][...],
##                                       vorticityNoG[0][...],
##                                       vorticityNoG[1][...],
##                                       vorticityNoG[2][...])

#        # Pure Vorticity diffusion
#        vorticityNoG[0][...], vorticityNoG[1][...], vorticityNoG[2][...] = \
#            fftw2py.solve_diffusion_3d(self.viscosity * dt,
#                                                  vorticityNoG[0][...],
#                                                  vorticityNoG[1][...],
#                                                  vorticityNoG[2][...])

#        for i in xrange(self.dim):
#            self.vorticity[i][ind0a:ind0b, ind1a:ind1b, ind2a:ind2b] = \
#                vorticityNoG[i][...]

        self.compute_time = time.time() - self.compute_time
        self.total_time += self.compute_time
        return self.compute_time

    def printComputeTime(self):
#        self.timings_info[0] = "\"Diffusion calculation total\""
#        self.timings_info[1] = str(self.total_time)
        print "Time of the last diffusion calculation loop :", \
              self.compute_time

    def __str__(self):
        s = "Diffusion_d (DiscreteOperator). " + DiscreteOperator.__str__(self)
        return s

if __name__ == "__main__":
    print __doc__
    print "- Provided class : Diffusion_d"
    print Diffusion_d.__doc__
