"""
@package parmepy.operator.discrete

Discrete operator representation.
"""
from abc import ABCMeta, abstractmethod
from parmepy.constants import debug


class DiscreteOperator(object):
    """
    Abstract description of a discretized operator.
    """

    __metaclass__ = ABCMeta

    @debug
    def __new__(cls, *args, **kw):
        return object.__new__(cls, *args, **kw)

    @debug
    @abstractmethod
    def __init__(self, variables, method=None):
        """
        Create an empty discrete operator.
        """
        ## Input variables
        self.input = []
        ## Output variables
        self.output = []
        ## variables
        self.variables = variables
        ## Operator numerical method.
        self.method = method
        self.numMethod = None
        ## Total compute time
        self.total_time = 0.
        ## Operator name
        self.name = "?"
        ## Timings informations
        self.time_info = ["", ""]

    @abstractmethod
    def printComputeTime(self):
        """Print total computing time."""
        raise NotImplementedError("Need to override method in " +
                                  "a subclass of DiscreteOperator")

    @debug
    @abstractmethod
    def apply(self, t, dt, ite):
        """
        Abstract method, apply operaton on a variable.
        Must be implemented by sub-class.
        """
        raise NotImplementedError("Need to override method in " +
                                  "a subclass of DiscreteOperator")

    @debug
    def finalize(self):
        """
        Method called on simulation finalization.
        Clean operator at time loop end.
        """
        pass

    def __str__(self):
        """ToString method"""
        s = "Applying on following fields : \n"
        if self.input is not None:
            s += "Input fields : \n"
            for f in self.input:
                s += str(f) + "\n"
        if self.output is not None:
            s += "Output fields : \n"
            for f in self.output:
                s += str(f) + "\n"
        return s

if __name__ == "__main__":
    print __doc__
    print "- Provided class : DiscreteOperator"
    print DiscreteOperator.__doc__
