# -*- coding: utf-8 -*-
"""
@package physics
Compute Energy and Enstrophy
"""
from parmepy.constants import np
import mpi4py.MPI as MPI
from parmepy.operator.monitors.monitoring import Monitoring
from parmepy.mpi import main_comm, main_rank, MPI
import time


class Energy_enstrophy(Monitoring):
    """
    Compute the forces according the Noca s formula
    """

    def __init__(self, fields=[],
                 frequency=None, outputPrefix=None):
        """
        Constructor.
        @param fields : List of fields
        @param frequency : output file producing frequency.
        @param outputPrefix : file name prefix, contains relative path.
        """
        Monitoring.__init__(self, fields, frequency)
        self.outputPrefix = outputPrefix
        self.velocity = None
        self.vorticity = None
        for f in fields:
            if f.name == 'Velocity':
                self.velocity = f
            if f.name == 'Vorticity':
                self.vorticity = f
        if self.velocity == None or self.vorticity == None:
            raise ValueError("Missing velocity or vorticity field")

        self.bufferVelo = 0.
        if (main_rank == 0):
            self.f = open(self.outputPrefix, 'w')

    def apply(self, t, dt, ite):
        """
        Computation of kinetic energy, enstrophy &
        Checking energy and enstrophy decay (--> is(recoveredViscosity == viscosity) ?)
        """
        velo = self.velocity.discreteField[0]
        vort = self.vorticity.discreteField[0]

        ind0a = velo.topology.mesh.local_start[0]
        ind0b = velo.topology.mesh.local_end[0] + 1
        if (velo.dimension > 1):
            ind1a = velo.topology.mesh.local_start[1]
            ind1b = velo.topology.mesh.local_end[1] + 1
        if (velo.dimension > 2):
            ind2a = velo.topology.mesh.local_start[2]
            ind2b = velo.topology.mesh.local_end[2] + 1

        dvol = np.prod(velo.topology.mesh.space_step)

        squareNormVel = np.sum(velo[0][ind0a:ind0b, ind1a:ind1b, ind2a:ind2b] ** 2 + \
                        velo[1][ind0a:ind0b, ind1a:ind1b, ind2a:ind2b] ** 2 + \
                        velo[2][ind0a:ind0b, ind1a:ind1b, ind2a:ind2b] ** 2) * dvol

        energy = 0.5 * squareNormVel

        ind0a = vort.topology.mesh.local_start[0]
        ind0b = vort.topology.mesh.local_end[0] + 1
        if (vort.dimension > 1):
            ind1a = vort.topology.mesh.local_start[1]
            ind1b = vort.topology.mesh.local_end[1] + 1
        if (vort.dimension > 2):
            ind2a = vort.topology.mesh.local_start[2]
            ind2b = vort.topology.mesh.local_end[2] + 1

        dvol = np.prod(vort.topology.mesh.space_step)

        enstrophy = np.sum(vort[0][ind0a:ind0b, ind1a:ind1b, ind2a:ind2b] ** 2 + \
                        vort[1][ind0a:ind0b, ind1a:ind1b, ind2a:ind2b] ** 2 + \
                        vort[2][ind0a:ind0b, ind1a:ind1b, ind2a:ind2b] ** 2) * dvol

        recoveredViscosity = - (0.5 / (dt * enstrophy)) * (squareNormVel -  self.bufferVelo)

        self.bufferVelo = squareNormVel

        if ((main_rank == 0) and (ite % self.freq == 0) and (t > dt)):
            self.f.write("%s   %s   %s   %s\n" % (t, energy, enstrophy, recoveredViscosity))

    def __str__(self):
        s = "Energy_enstrophy. "
        return s

if __name__ == "__main__":
    print __doc__
    print "- Provided class : Energy_enstrophy"
    print Energy_enstrophy.__doc__
