"""
@package parmepy.operator.gpu_particle_advection_1k

Discrete advection representation
"""
from abc import ABCMeta, abstractmethod
from parmepy import __VERBOSE__
from parmepy.constants import np, debug
from parmepy.operator.particle_advection import ParticleAdvection
from parmepy.fields.continuous import Field
from parmepy.fields.gpu_discrete import GPUVectorField
from parmepy.fields.gpu_discrete import GPUScalarField
from parmepy.gpu import PARMES_REAL_GPU, PARMES_DOUBLE_GPU, \
    PARMES_INTEGER_GPU, GPU_SRC, cl
from parmepy.gpu.tools import get_opencl_environment, \
    _check_double_precision_capability, _set_WorkItems, _get_precision,\
    build_src
from parmepy.numerics.gpu_kernel import KernelLauncher
from parmepy.numerics.gpu_kernel import KernelListLauncher
from parmepy.numerics.splitting import Splitting


class GPUParticleAdvection(ParticleAdvection):
    """
    Particle advection operator representation on GPU.

    """
    __metaclass__ = ABCMeta

    @debug
    @abstractmethod
    def __init__(self, advec,
                 platform_id=0, device_id=0,
                 device_type='gpu',
                 method='', src=None, precision=PARMES_REAL_GPU,
                 splittingConfig='o2'):
        """
        Create a Advection operator.
        Work on a given scalar at a given velocity to produce scalar
        distribution at new positions.

        @param advec : Transport operator
        @param idVelocityD : Index of velocity discretisation to use.
        @param idScalarD : Index of scalar discretisation to use.
        @param method : the method to use.
        """
        ParticleAdvection.__init__(self, advec)
        self.name = "advection_gpu"
        self.method = method
        self.user_gpu_src = src
        self.num_method = None
        self.splittingConfig = splittingConfig
        self.resolution = self.advectedFields.topology.mesh.resolution
        self.workItemNumber = None
        self.platform, self.device, self.ctx, self.queue = \
            get_opencl_environment(platform_id, device_id, device_type,
                                   precision)
        self.warning = True

    @debug
    def setUp(self):
        self.gpu_precision = _get_precision()
        self.workItemNumber, self.gwi, self.lwi = _set_WorkItems(
            self.resolution)
        if __VERBOSE__:
            print "Work-items basic config : ",
            print self.workItemNumber, self.gwi, self.lwi
        dim = self.advectedFields.topology.dim
        self.coord_min = np.ones(4, dtype=self.gpu_precision)
        self.mesh_size = np.ones(4, dtype=self.gpu_precision)
        self.coord_min[0:dim] = np.asarray(self.advectedFields.topology.mesh.origin,
                                           dtype=self.gpu_precision)
        self.mesh_size[0:dim] = np.asarray(
            self.advectedFields.topology.mesh.space_step,
            dtype=self.gpu_precision)
        if __VERBOSE__:
            print "=== Kernel sources ==="
        self.build_options = ""
        if _check_double_precision_capability(self.device,
                                              self.gpu_precision):
            self.build_options += " -cl-fp32-correctly-rounded-divide-sqrt "
        if self.gpu_precision is PARMES_REAL_GPU:
            self.build_options += " -cl-single-precision-constant"
        self.build_options += " -D WIDTH=" + str(self.resolution[0])
        self.build_options += " -D WGN=" + str(self.workItemNumber)
        self.gpu_src = self._collect_usr_cl_src(self.user_gpu_src)
        self.gpu_src += self._collect_kernels_cl_src()
        if __VERBOSE__:
            print "===\n"
        ## Build code
        self.prg = build_src(self.device, self.ctx, self.gpu_src,
                             self.build_options, self.gpu_precision)
        print "=== OpenCL Buffer allocations ==="
        self._buffer_allocations()
        print "===\n"
        print "=== OpenCL Buffer initialisation ==="
        self._buffer_initialisations()
        print "===\n"
        self._kernel_affectations()
        self.numMethod = Splitting(self._apply_in_dir,
                                   self.advectedFields.dimension,
                                   config=self.splittingConfig)

    def _buffer_initialisations(self):
        data, transfer_time, compute_time = 0, 0., 0.
        for gpudf in self.variables:
            match = 'init' + gpudf.name
            initKernel = None
            # Looking for initKernel
            for k in self.prg.all_kernels():
                k_name = k.get_info(cl.kernel_info.FUNCTION_NAME)
                if match.find(k_name) >= 0:
                    gpudf.initialization_kernel = KernelLauncher(
                        cl.Kernel(self.prg, k_name), self.queue,
                        self.gwi, self.lwi)

    def _collect_kernels_cl_src_copy(self):
        gpu_src = ""
        ## copy
        if len(self.resolution) == 3:
            if self.gpu_precision is PARMES_REAL_GPU:
                src_copy = 'copy_3D_opt4.cl'
                self.build_options += " -D TILE_DIM_COPY=16"
                self.build_options += " -D BLOCK_ROWS_COPY=8"
                self.copy_gwi = (int(self.resolution[0] / 4),
                                 int(self.resolution[1] / 2),
                                 int(self.resolution[2]))
                self.copy_lwi = (4, 8, 1)
            else:
                src_copy = 'copy_3D_locMem.cl'
                self.build_options += " -D TILE_DIM_COPY=32"
                self.build_options += " -D BLOCK_ROWS_COPY=8"
                self.copy_gwi = (int(self.resolution[0]),
                                 int(self.resolution[1] / 4),
                                 int(self.resolution[2]))
                self.copy_lwi = (32, 8, 1)
        else:
            if self.gpu_precision is PARMES_REAL_GPU:
                src_copy = 'copy_2D_opt2.cl'
                self.build_options += " -D TILE_DIM_COPY=16"
                self.build_options += " -D BLOCK_ROWS_COPY=8"
                self.copy_gwi = (int(self.resolution[0] / 2),
                                 int(self.resolution[1] / 2))
                self.copy_lwi = (8, 8)
            else:
                src_copy = 'copy_2D_opt2.cl'
                self.build_options += " -D TILE_DIM_COPY=32"
                self.build_options += " -D BLOCK_ROWS_COPY=2"
                self.copy_gwi = (int(self.resolution[0] / 2),
                                 int(self.resolution[1] / 16))
                self.copy_lwi = (16, 2)
        f = open(GPU_SRC + src_copy)
        gpu_src += "".join(f.readlines())
        f.close()
        if __VERBOSE__:
            print  "   - copy:", GPU_SRC + src_copy
            print "       - config :", self.copy_gwi, self.copy_lwi
        return gpu_src

    def _collect_kernels_cl_src_transpositions(self):
        gpu_src = ""
        ## transpose
        if len(self.resolution) == 3:
            if self.gpu_precision is PARMES_REAL_GPU:
                src_transpose_xy = \
                    'transpose_3D_xy_coalesced_locPad_Diag_2Dslice_opt2.cl'
                self.build_options += " -D TILE_DIM_XY=16 -D BLOCK_ROWS_XY=8"
                self.transpose_xy_gwi = (int(self.resolution[0] / 2),
                                         int(self.resolution[1] / 2),
                                         int(self.resolution[2]))
                self.transpose_xy_lwi = (8, 8, 1)
                src_transpose_xz = \
                    'transpose_3D_xz_coalesced_locPad_Diag_bis_3DBlock.cl'
                self.build_options += " -D TILE_DIM_XZ=16 -D BLOCK_ROWS_XZ=4"
                self.transpose_xz_gwi = (int(self.resolution[0]),
                                         int(self.resolution[1] / 4),
                                         int(self.resolution[2] / 4))
                self.transpose_xz_lwi = (16, 4, 4)
            else:
                src_transpose_xy = \
                    'transpose_3D_xy_coalesced_locPad_Diag_2Dslice_opt4.cl'
                self.build_options += " -D TILE_DIM_XY=32 -D BLOCK_ROWS_XY=4"
                self.transpose_xy_gwi = (int(self.resolution[0] / 4),
                                         int(self.resolution[1] / 8),
                                         int(self.resolution[2]))
                self.transpose_xy_lwi = (8, 4, 1)
                src_transpose_xz = \
                    'transpose_3D_xz_coalesced_Diag_bis_3DBlock.cl'
                self.build_options += " -D TILE_DIM_XZ=8 -D BLOCK_ROWS_XZ=2"
                self.transpose_xz_gwi = (int(self.resolution[0]),
                                         int(self.resolution[1] / 4),
                                         int(self.resolution[2] / 4))
                self.transpose_xz_lwi = (8, 2, 2)
        else:
            if self.gpu_precision is PARMES_REAL_GPU:
                src_transpose_xy = \
                    'transpose_2D_xy_coalesced_locPad_Diag_opt4.cl'
                self.build_options += " -D TILE_DIM_XY=32 -D BLOCK_ROWS_XY=8"
                self.transpose_xy_gwi = (int(self.resolution[0] / 4),
                                         int(self.resolution[1]) / 4)
                self.transpose_xy_lwi = (8, 8)
            else:
                src_transpose_xy = \
                    'transpose_2D_xy_coalesced_locPad_Diag_opt4.cl'
                self.build_options += " -D TILE_DIM_XY=32 -D BLOCK_ROWS_XY=2"
                self.transpose_xy_gwi = (int(self.resolution[0] / 4),
                                         int(self.resolution[1]) / 16)
                self.transpose_xy_lwi = (8, 2)
        f = open(GPU_SRC + src_transpose_xy)
        gpu_src += "".join(f.readlines())
        f.close()
        if len(self.resolution) == 3:
            f = open(GPU_SRC + src_transpose_xz)
            gpu_src += "".join(f.readlines())
            f.close()
        if __VERBOSE__:
            print  "   - transpose_xy:", GPU_SRC + src_transpose_xy
            print "       - config :", self.transpose_xy_gwi,
            print self.transpose_xy_lwi
            if len(self.resolution) == 3:
                print  "   - transpose_xz:", GPU_SRC + src_transpose_xz
                print "       - config :",
                print self.transpose_xz_gwi, self.transpose_xz_lwi
        return gpu_src

    def _collect_usr_cl_src(self, usr_src):
        gpu_src = ""
        if __VERBOSE__:
            print "Sources files (user): ", usr_src
        if usr_src is not None:
            if isinstance(usr_src, list):
                for src in usr_src:
                    f = open(src, 'r')
                    gpu_src += "".join(f.readlines())
                    f.close()
            else:
                f = open(usr_src, 'r')
                gpu_src += "".join(f.readlines())
                f.close()
        return gpu_src

    @debug
    def _apply_in_dir(self, t, dt, d, split_id):
        ## Check new data on host
        for df in self.input:
            if df.contains_data:
                df.toDevice()
        dim = self.advectedFields.dimension
        self.queue.finish()
        # Particle init
        if (split_id == 0) or ((split_id == dim) and (d == dim - 1)):
            evt_init = self.init_copy(self.advectedFields.gpu_data,
                                      self.part_scalar.gpu_data)
        else:
            if dim == 2:
                evt_init = \
                    self.init_transpose(self.advectedFields.gpu_data,
                                        self.part_scalar.gpu_data)
            else:
                if split_id == d:
                    evt_init = \
                        self.init_transpose(d - 1,
                                            self.advectedFields.gpu_data,
                                            self.part_scalar.gpu_data)
                else:
                    evt_init = \
                        self.init_transpose(d, self.advectedFields.gpu_data,
                                            self.part_scalar.gpu_data)
        return evt_init

    @debug
    def apply(self, t, dt, ite):
        self.numMethod(t, dt)

    @debug
    def finalize(self):
        """
        Free OpenCL memory.
        """
        for gpudf in self.variables:
            if __VERBOSE__:
                print "deallocate :", gpudf.name
            if gpudf.vector:
                [gpudf.gpu_data[d].release()
                 for d in xrange(len(self.resolution))]
            else:
                gpudf.gpu_data.release()

    def __str__(self):
        s = "Advection_P (DiscreteOperator). " + DiscreteOperator.__str__(self)
        return s

if __name__ == "__main__":
    print __doc__
    print "- Provided class : Transport_d"
    print AdvectionDOp.__doc__
