"""
@package parmepy.operator.gpu_particle_advection_1k

Discrete advection representation
"""
from parmepy.constants import np, debug
from parmepy import __VERBOSE__
from parmepy.operator.gpu_particle_advection import GPUParticleAdvection
from parmepy.fields.continuous import Field
from parmepy.fields.gpu_discrete import GPUVectorField
from parmepy.fields.gpu_discrete import GPUScalarField
from parmepy.gpu import PARMES_REAL_GPU, PARMES_DOUBLE_GPU, \
    PARMES_INTEGER_GPU, GPU_SRC, cl
from parmepy.gpu.tools import get_opencl_environment, \
    _check_double_precision_capability, _set_WorkItems
from parmepy.numerics.gpu_kernel import KernelLauncher
from parmepy.numerics.gpu_kernel import KernelListLauncher
from parmepy.numerics.splitting import Splitting


class GPUParticleAdvection1k(GPUParticleAdvection):
    """
    Particle advection operator representation on GPU.

    """

    @debug
    def __init__(self, advec,
                 platform_id=0, device_id=0,
                 device_type='gpu',
                 method='', src=None, precision=PARMES_REAL_GPU,
                 splittingConfig='o2'):
        """
        Create a Advection operator.
        Work on a given scalar at a given velocity to produce scalar
        distribution at new positions.

        @param advec : Transport operator
        @param idVelocityD : Index of velocity discretisation to use.
        @param idScalarD : Index of scalar discretisation to use.
        @param method : the method to use.
        """
        GPUParticleAdvection.__init__(self, advec,
                                      platform_id, device_id,
                                      device_type,
                                      method, src,
                                      precision, splittingConfig)

    @debug
    def setUp(self):
        GPUParticleAdvection.setUp(self)

    def _kernel_affectations(self):
        self.num_advec_and_remesh = KernelLauncher(
            self.prg.advection_and_remeshing,
            self.queue,
            self.gwi_advec_and_remesh,
            self.lwi_advec_and_remesh)
        self.init_copy = KernelLauncher(self.prg.copy,
                                        self.queue,
                                        self.copy_gwi,
                                        self.copy_lwi)
        if len(self.resolution) == 3:
            self.init_transpose = \
                KernelListLauncher([self.prg.transpose_xy,
                                    self.prg.transpose_xz],
                                   self.queue,
                                   [self.transpose_xy_gwi,
                                    self.transpose_xz_gwi],
                                   [self.transpose_xy_lwi,
                                    self.transpose_xz_lwi])
        else:
            self.init_transpose = \
                KernelLauncher(self.prg.transpose_xy,
                               self.queue,
                               self.transpose_xy_gwi,
                               self.transpose_xy_lwi)

    def _buffer_allocations(self):
        ## Velocity.
        GPUVectorField.fromVectorField(self.queue, self.velocity,
                                       self.gpu_precision)
        ## Transported scalar.
        GPUScalarField.fromScalarField(self.queue, self.advectedFields,
                                       self.gpu_precision)
        ## Result scalar
        particle_scalar = Field(self.advectedFields.topology.domain,
                                "Particle_Scalar",
                                vector=False)
        self.part_scalar, dfId = particle_scalar.discretize(
            self.advectedFields.topology)
        GPUScalarField.fromScalarField(
            self.queue, self.part_scalar,
            self.gpu_precision)
        self.variables = [self.advectedFields, self.velocity,
                          self.part_scalar]
        total_mem_used = 0
        for gpuv in self.variables:
            total_mem_used += gpuv.mem_size
        print ""
        print "Total Global Memory used : ", total_mem_used,
        print "Bytes (", total_mem_used / (1024 ** 2), "MB)",
        print "({0:.3f} %)".format(
            100 * total_mem_used / (self.device.global_mem_size * 1.))

    def _collect_kernels_cl_src(self):
        gpu_src = ""
        ## advection and remeshing
        advec_and_remesh_vector_width = 4
        src_advec_and_remesh = 'advection_and_remesh_'
        if len(self.resolution) == 3:
            if self.gpu_precision is PARMES_REAL_GPU:
                if self.method.find('m4prime') >= 0:
                    src_advec_and_remesh += \
                        'm4prime_3D_opt4_builtin_private4.cl'
                    src_remesh_weights = 'weights_m4prime_opt4_builtin.cl'
                elif self.method.find('m6prime') >= 0:
                    src_advec_and_remesh += '3D_opt4_builtin_private4.cl'
                    src_remesh_weights = 'weights_m6prime_opt4_builtin.cl'
                elif self.method.find('m8prime') >= 0:
                    src_advec_and_remesh += \
                        'm8prime_3D_opt4_builtin_private4.cl'
                    src_remesh_weights = 'weights_m8prime_opt4_builtin.cl'
                else:
                    print 'No remeshing fromula specified, use default m6prime'
                    src_advec_and_remesh += '3D_opt4_builtin_private4.cl'
                    src_remesh_weights = 'weights_m6prime_opt4_builtin.cl'
            else:
                if self.method.find('m4prime') >= 0:
                    src_advec_and_remesh += \
                        'm4prime_3D_opt4_builtin_private4_noBC.cl'
                    src_remesh_weights = 'weights_m4prime_opt4_builtin.cl'
                elif self.method.find('m6prime') >= 0:
                    src_advec_and_remesh += '3D_opt4_builtin_private4_noBC.cl'
                    src_remesh_weights = 'weights_m6prime_opt4_builtin.cl'
                elif self.method.find('m8prime') >= 0:
                    src_advec_and_remesh += \
                        'm8prime_3D_opt4_builtin_private4_noBC.cl'
                    src_remesh_weights = 'weights_m8prime_opt4_builtin.cl'
                else:
                    print 'No remeshing fromula specified, use default m6prime'
                    src_advec_and_remesh += '3D_opt4_builtin_private4_noBC.cl'
                    src_remesh_weights = 'weights_m6prime_opt4_builtin.cl'
        else:
            if self.gpu_precision is PARMES_REAL_GPU:
                advec_and_remesh_vector_width = 8
                if self.method.find('m4prime') >= 0:
                    src_advec_and_remesh += 'm4prime_2D_opt8_builtin_noBC.cl'
                    src_remesh_weights = 'weights_m4prime_builtin.cl'
                elif self.method.find('m6prime') >= 0:
                    src_advec_and_remesh += '2D_opt8_builtin_noBC.cl'
                    src_remesh_weights = 'weights_m6prime_builtin.cl'
                elif self.method.find('m8prime') >= 0:
                    src_advec_and_remesh += 'm8prime_2D_opt8_builtin_noBC.cl'
                    src_remesh_weights = 'weights_m8prime_builtin.cl'
                else:
                    print 'No remeshing fromula specified, use default m6prime'
                    src_advec_and_remesh += '2D_opt8_builtin_noBC.cl'
                    src_remesh_weights = 'weights_m6prime_builtin.cl'
            else:
                if self.method.find('m4prime') >= 0:
                    src_advec_and_remesh += 'm4prime_2D_opt4_builtin_noBC.cl'
                    src_remesh_weights = 'weights_m4prime_builtin.cl'
                elif self.method.find('m6prime') >= 0:
                    src_advec_and_remesh += '2D_opt4_builtin_noBC.cl'
                    src_remesh_weights = 'weights_m6prime_builtin.cl'
                elif self.method.find('m8prime') >= 0:
                    src_advec_and_remesh += 'm8prime_2D_opt4_builtin_noBC.cl'
                    src_remesh_weights = 'weights_m8prime_builtin.cl'
                else:
                    print 'No remeshing fromula specified, use default m6prime'
                    src_advec_and_remesh += '2D_opt4_builtin_noBC.cl'
                    src_remesh_weights = 'weights_m6prime_builtin.cl'
        self.WINb_advec_and_remesh, self.gwi_advec_and_remesh, \
            self.lwi_advec_and_remesh = _set_WorkItems(
                self.resolution, advec_and_remesh_vector_width)
        self.build_options += " -D WGNADVECANDREMESH="
        self.build_options += str(self.WINb_advec_and_remesh)
        f = open(GPU_SRC + src_remesh_weights)
        gpu_src += "".join(f.readlines())
        f.close()
        f = open(GPU_SRC + src_advec_and_remesh)
        gpu_src += "".join(f.readlines())
        f.close()
        gpu_src += self._collect_kernels_cl_src_copy()
        gpu_src += self._collect_kernels_cl_src_transpositions()
        if __VERBOSE__:
            print  "   - remeshing weights:", GPU_SRC + src_remesh_weights
            print  "   - advection and remeshing:",
            print GPU_SRC + src_advec_and_remesh
            print "       - config :",
            print self.gwi_advec_and_remesh, self.lwi_advec_and_remesh
        return gpu_src

    @debug
    def _apply_in_dir(self, t, dt, d, split_id):
        """
        Apply advection operator.

        @param t : current time.
        @param dt : time step.
        @param d : Direction of splitting.

        Advection algorithm:
        @li 1. Particle initialization : \n
                 - by copy scalar from grid to particles if previous splitting
                 direction equals current splitting direction.\n
                 - by transposition of scalar from grid to particle.
        @li 2. Particle advection :\n
                 - compute particle position in splitting direction as a
                 scalar. Performs a RK2 resolution of dx_p/dt = a_p.
        @li 3. Profile timings of OpenCL kernels.
        """
        evt_init = GPUParticleAdvection._apply_in_dir(self, t, dt, d, split_id)
        # Advection
        evt_advec = self.num_advec_and_remesh(self.velocity.gpu_data[d],
                                              self.part_scalar.gpu_data,
                                              self.advectedFields.gpu_data,
                                              self.gpu_precision(dt),
                                              self.coord_min[d],
                                              self.mesh_size[d])
        for df in self.output:
            df.data_on_device = True
            df.contains_data = False
        # Get timpings from OpenCL events
        self.queue.finish()
        c_time_init = (evt_init.profile.end - evt_init.profile.start) * 1e-9
        c_time_advec = (evt_advec.profile.end - evt_advec.profile.start) * 1e-9
        self.compute_time[d] += c_time_advec
        if split_id == 0:
            self.compute_time_copy[d] += c_time_init
        else:
            if self.advectedFields.topology.dim == 2:
                self.compute_time_swap[0] += c_time_init
            else:
                if min(split_id, d) == 0:
                    self.compute_time_swap[0] += c_time_init
                else:
                    self.compute_time_swap[1] += c_time_init
        self.total_time += (c_time_advec + c_time_init)
        return (c_time_advec + c_time_init)

    def printComputeTime(self):
        self.time_info[0] = "\"Advection total\" "
        self.time_info[0] += "\"copy\" \"swap xy\" \"swap xz\" "
        self.time_info[0] += "\"Advection x\" \"Advection y\" \"Advection z\" "
        self.time_info[1] = str(self.total_time) + "  "
        self.time_info[1] += str(self.compute_time_copy[0]) + "  "
        self.time_info[1] += str(self.compute_time_swap[0]) + "  "
        self.time_info[1] += str(self.compute_time_swap[1]) + "  "
        self.time_info[1] += str(self.compute_time[0]) + "  "
        self.time_info[1] += str(self.compute_time[1]) + "  "
        self.time_info[1] += str(self.compute_time[2]) + "  "
        print "Advection total time : ", self.total_time, self.call_number
        print "\t Advection init copy :", self.compute_time_copy,
        print self.init_copy.call_number
        print "\t Advection init swap :", self.compute_time_swap,
        print self.init_transpose.call_number
        print "\t Advection :", self.compute_time,
        print self.num_advec_and_remesh.call_number

    def __str__(self):
        s = "Advection_P (DiscreteOperator). " + DiscreteOperator.__str__(self)
        return s

if __name__ == "__main__":
    print __doc__
    print "- Provided class : Transport_d"
    print AdvectionDOp.__doc__
