"""
@package parmepy.tools.printer

Classes for handling ouputs.
"""
from parmepy.constants import np, S_DIR, PARMES_REAL
from parmepy.operator.monitors.monitoring import Monitoring
import evtk.hl as evtk
import time


class Printer(Monitoring):
    """
    Basic printer.

    Performs outputs in VTK images.
    """

    def __init__(self, frequency=0, fields=[], outputPrefix='./out_'):
        """
        Create a results printer for given fields, filename
        prefix (relative path) and an output frequency.

        @param fields : Fields to output.
        @param frequency : output file producing frequency.
        @param outputPrefix : file name prefix, contains relative path.
        """
        Monitoring.__init__(self, fields, frequency)
        ## Filename prefix
        self.outputPrefix = outputPrefix
        ## Method to collect data in case of distributed data
        self.get_data_method = None
        if self.freq != 0:
            ## Printer core method
            self.step = self._printStep
        else:
            self.step = self._passStep
        ## Printing compute time
        self.compute_time = 0.
        self.call_number = 0

    def apply(self, t, dt, ite):
        self.step(ite)

    def _build_vtk_dict(self):
        """Build a dictionary from fields to VTK image."""
        res = {}
        for f in self.variables:
            for df in f.discreteField:
                if f.vector:
                    for d in xrange(len(df.data)):
                        if len(df.data[d].shape) == 2:
                            res[df.name + S_DIR[d]] = np.expand_dims(
                                df.data[d], axis=2).astype(PARMES_REAL)
                        else:
                            res[df.name + S_DIR[d]] = \
                                df.data[d].astype(PARMES_REAL)
                else:
                    if len(df.data.shape) == 2:
                        res[df.name] = \
                            np.expand_dims(df.data, axis=2).astype(PARMES_REAL)
                    else:
                        res[df.name] = df.data.astype(PARMES_REAL)
        return res

    def set_get_data_method(self, method):
        """Set data collect method."""
        if callable(method):
            self.get_data_method = method
        else:
            raise ValueError("Cannot set non callable method to get data " +
                             "to print. Given method : " + str(method))

    def _passStep(self, ite):
        """A code method that do nothing."""
        pass

    def _printStep(self, ite):
        """
        Printer core method.

        Write fields data to file on necessary.\n
        VTK images are writed. If fails, classical ascii output is performed.
        """
        t = time.time()
        if (ite % self.freq) == 0:
            self.call_number += 1
            print "== IO"
            for f in self.variables:
                for df in f.discreteField:
                    df.toHost()
            filename = self.outputPrefix + "results_{0:05d}.dat".format(ite)
            print "Print file : " + filename
            ## VTK output \todo: Need fix in 2D, getting an IOError.
            try:
                evtk.imageToVTK(filename, pointData=self._build_vtk_dict())
            except IOError:
                if self.variables[0].domain.dimension == 2:
                    ## Standard output
                    f = open(filename, 'w')
                    shape = self.variables[0].discreteField[0].topology.mesh.resolution
                    coords = self.variables[0].discreteField[0].topology.mesh.coords
                    if len(shape) == 2:
                        for i in xrange(shape[0] - 1):
                            for j in xrange(shape[1] - 1):
                                f.write("{0:8.12} {1:8.12} ".format(
                                        coords[0][i, 0], coords[1][0, j]))
                                for field in self.variables:
                                    if field.vector:
                                        f.write("{0:8.12} {1:8.12} ".format(
                                                field.discreteField[0][0][i, j],
                                                field.discreteField[0][1][i, j]))
                                    else:
                                        f.write("{0:8.12} ".format(
                                                field.discreteField[0][i, j]))
                                f.write("\n")
                    else:
                        for i in xrange(shape[0] - 1):
                            for j in xrange(shape[1] - 1):
                                for k in xrange(shape[2] - 1):
                                    f.write("{0:8.12} {1:8.12} {2:8.12} ".format(
                                            coords[0][i, 0, 0],
                                            coords[1][0, j, 0],
                                            coords[2][0, 0, k]))
                                    for field in self.variables:
                                        if field.vector:
                                            f.write("{0:8.12} {1:8.12} {2:8.12} ".format(
                                                    field.discreteField[0][0][i, j, k],
                                                    field.discreteField[0][1][i, j, k],
                                                    field.discreteField[0][2][i, j, k]))
                                        else:
                                            f.write("{0:8.12} ".format(
                                                    field.discreteField[0][i, j, k]))
                                    f.write("\n")
                    f.close()
            print "==\n"
        self.compute_time += (time.time() - t)

    def printComputeTime(self):
        self.time_info[0] = "\"IO\" "
        self.time_info[1] = str(self.compute_time) + "  "
        print "IO total time : ", self.compute_time, self.call_number


if __name__ == "__main__":
    print __doc__
    print "- Provided class : " + providedClass
    print eval(providedClass).__doc__
