# -*- coding: utf-8 -*-
"""
@package parmepy.operator.multiphase

MultiPhase Rot Grad P
"""
from parmepy.operator.continuous import Operator
from parmepy.operator.multiphase_d import Pressure_d
from stretching_d import Stretching_d


class Pressure(Operator):
    """
    Pressure operator representation
    """

    @debug
    def __init__(self, velocity, vorticity, viscosity, density,
                 resolutions=None, method='',
                 **other_config):
        """
        Constructor.
        Create a Pressure operator from given velocity variables.

        @param velocity ContinuousVectorField : velocity variable.
        """
        Operator.__init__(self, [velocity, vorticity, density])
        ## velocity variable (vector)
        self.velocity = velocity
        self.vorticity = vorticity
        self.density = density
        self.viscosity = viscosity
        self.resolutions = resolutions
        self.method = method
        self.config = other_config

    @debug
    def setUp(self):
        """
        Pressure operator discretization method.
        Create a discrete Pressure operator from given specifications.

        @param idVelocityD : Index of velocity discretisation to use.
        @param idVelocityD : Index of vorticity discretisation to update.
        @param idDensityD : Index of density discretisation to use.
        @param propertyOp : choice of the property garantied by the stretching Op ( divConservation or massConservation)
        @param method : the method to use. (Euler, RK, ..) default : RK3
        """
        Operator.setUp(self)
        if self.resolutions is None:
            raise ValueError("Advection setUp: no resolution given.")
        #variables discretization
        for v in self.variables:
            v.discretize(self.resolutions[v])
        self.discreteOperator = Pressure_d(self, method=self.method,
                                           **self.config)
        self.discreteOperator.setUp()


    def __str__(self):
        """ToString method"""
        s = " Pressure operator (ContinuousOperator)"
        if self.discreteOperator is not None:
            s += "with the following discretization:\n"
            s += str(self.discreteOperator)
        else:
            s += ". Not discretised"
        return s + "\n"

if __name__ == "__main__":
    print __doc__
    print "- Provided class : Pressure"
    print Stretching.__doc__
