"""
@package parmepy.operator.particle_advection

Discrete advection representation
"""
from parmepy.constants import np, debug
from parmepy.operator.discrete import DiscreteOperator
from parmepy.fields.continuous import Field
from parmepy.gpu import PARMES_REAL_GPU
from parmepy.numerics.interpolation import Linear
from parmepy.numerics.splitting import Splitting
import time


class ParticleAdvection(DiscreteOperator):
    """
    Particle advection operator representation.

    """

    @debug
    def __init__(self, advec, method=''):
        """
        Create a Advection operator.
        Work on a given advectedFields at a given velocity
        to produce advectedFields
        distribution at new positions.

        @param advec : Transport operator
        @param method : the method to use.
        """
        self.advec = advec
        ## Velocity.
        self.velocity = self.advec.velocity.discreteField[
            self.advec.discreteFieldId[self.advec.velocity]]
        ## Advected Fields.
        self.advectedFields = self.advec.advectedFields.discreteField[
            self.advec.discreteFieldId[self.advec.advectedFields]]
        DiscreteOperator.__init__(self,
                                  [self.velocity, self.advectedFields],
                                  method)
        self.input = [self.velocity, self.advectedFields]
        self.output = [self.advectedFields]
        ## Compute time detailed per directions
        self.compute_time = [0., 0., 0.]
        ## Compute time for copy detailed per directions
        self.compute_time_copy = [0., 0., 0.]
        ## Compute time for transposition detailed per directions
        self.compute_time_swap = [0., 0., 0.]
        self.call_number = [0, 0, 0]
        self.name = "advection_default"
        self.num_method = None

    @debug
    def setUp(self):
        ## Result position
        particle_position = Field(self.advec.advectedFields.domain,
                                  "Particle_Position",
                                  vector=False)
        particle_position.discretize(
            self.advec.resolutions[self.advec.advectedFields])
        self.part_position = particle_position.discreteField[0]
        ## Result advectedFields
        particle_advectedFields = Field(self.advec.advectedFields.domain,
                                        "Particle_AdvectedFields",
                                        vector=False)
        particle_advectedFields.discretize(
            self.advec.resolutions[self.advec.advectedFields])
        self.part_advectedFields = particle_advectedFields.discreteField[0]
        if self.method.find('rk2') >= 0:
            self.name += '_rk2'
            from parmepy.numerics.integrators.runge_kutta2 import RK2
            self.num_advec = RK2()
        else:
            self.name += '_rk2'
            from parmepy.numerics.integrators.runge_kutta2 import RK2
            self.num_advec = RK2()
        if self.method.find('m6prime') >= 0:
            self.name += '_m6prime'
            from parmepy.numerics.remeshing.m6prime import M6Prime
            self.num_remesh = M6Prime(self.advectedFields.dimension)
        else:
            self.name += '_m6prime'
            from parmepy.numerics.remeshing.m6prime import M6Prime
            self.num_remesh = M6Prime(self.advectedFields.dimension)
        self.interpolation = [Linear(self.velocity[d])
                              for d in xrange(self.advectedFields.dimension)]
        self.numMethod = Splitting(
            self._apply_in_dir, self.advectedFields.dimension)
        print "Method used:", self.name

    def _apply_in_dir(self, t, dt, d, split_id):
        """
        Apply advection operator.

        @param t : current time.
        @param dt : time step.
        @param d : Direction of splitting.

        Advection algorithm:
        @li 1. Particle initialization : \n
                 - by copy advectedFields from grid to particles
                 if previous splitting direction equals current
                 splitting direction.\n
                 - by transposition of advectedFields from grid to particle.
        @li 2. Particle advection :\n
                 - compute particle position in splitting direction as a
                 advectedFields. Performs a RK2 resolution of dx_p/dt = a_p.
        @li 3. Profile timings of OpenCL kernels.
        """
        self.part_advectedFields[...] = self.advectedFields[...]
        self.part_position[...] = self.advectedFields.topology.mesh.coords[d]
        self.part_position[...] = self.num_advec(self.interpolation[d],
                                                 t, dt,
                                                 self.part_position.data)
        self.advectedFields[...] = \
            self.num_remesh(self.part_position.data,
                            self.part_advectedFields.data, d)

    @debug
    def apply(self, t, dt, ite):
        self.numMethod(t, dt)


    def printComputeTime(self):
        pass
        # self.timings_info[0] = "\"Advection total\" \"copy\" \"swap xy\" \"swap xz\" \"Advection x\" \"Advection y\" \"Advection z\" "
        # self.timings_info[1] = str(self.total_time) + "  " + str(self.compute_time_copy[0]) + "  "
        # self.timings_info[1] += str(self.compute_time_swap[0]) + "  " + str(self.compute_time_swap[1]) + "  "
        # self.timings_info[1] += str(self.compute_time[0]) + "  " + str(self.compute_time[1]) + "  " + str(self.compute_time[2]) + "  "
        # print "Advection total time : ", self.total_time, self.call_number
        # print "\t Advection init copy :", self.compute_time_copy, self.init_copy.call_number
        # print "\t Advection init swap :", self.compute_time_swap, self.init_transpose.call_number
        # print "\t Advection :", self.compute_time, self.numMethod.call_number

    def __str__(self):
        s = "Advection_P (DiscreteOperator). " + DiscreteOperator.__str__(self)
        return s

if __name__ == "__main__":
    print __doc__
    print "- Provided class : Transport_d"
    print AdvectionDOp.__doc__
