# -*- coding: utf-8 -*-
"""
@package operator
Discrete penalization representation
"""
from parmepy.constants import debug
from parmepy.operator.discrete import DiscreteOperator
from parmepy.numerics.differentialOperator import DifferentialOperator
from parmepy.constants import np
import time


class Penalization_d(DiscreteOperator):
    """
    Penalization operator representation.
    DiscreteOperator.DiscreteOperator specialization.
    """

    @debug
    def __init__(self, penal, method=''):
        """
        Constructor.
        Create a Penalization operator on a given continuous domain.
        Work on a given velocity and penalize it.

        @param penal : Penalization operator.
        """
        DiscreteOperator.__init__(self, penal)
        ## Continous penalization operator
        self.penal = penal
        ## Velocity.
        self.velocity = self.penal.velocity.discreteField[
            self.penal.discreteFieldId[self.penal.velocity]]
        ## Vorticity.
        self.vorticity = self.penal.vorticity.discreteField[
            self.penal.discreteFieldId[self.penal.vorticity]]
        ## Obstacle
        self.obstacle = penal.obstacle
        self.obstacle.setUp()
        ## Method
        self.method = method

    @debug
    def setUp(self):
        self.obstacle.definedObst.setUp(self.velocity.topology)

    @debug
    def apply(self, t, dt, ite):

        self.compute_time = time.time()

        coef = 1.0 / (1.0 + dt * self.penal.lambd[:])

        b = self.obstacle.definedObst.chiBoundary
        s = self.obstacle.definedObst.chiSolid
        p = self.obstacle.definedObst.chiPorous

        if self.velocity.dimension == 2 :
            self.velocity[0][b[0], b[1]] = coef[0] * self.velocity[0][b[0], b[1]]
            self.velocity[1][b[0], b[1]] = coef[0] * self.velocity[1][b[0], b[1]]

            self.velocity[0][s[0], s[1]] = coef[0] * self.velocity[0][s[0], s[1]]
            self.velocity[1][s[0], s[1]] = coef[0] * self.velocity[1][s[0], s[1]]

            self.velocity[0][p[0], p[1]] = coef[1] * self.velocity[0][p[0], p[1]]
            self.velocity[1][p[0], p[1]] = coef[1] * self.velocity[1][p[0], p[1]]

        if self.velocity.dimension == 3 :
            self.velocity[0][b[0], b[1], b[2]] = coef[0] * self.velocity[0][b[0], b[1], b[2]]
            self.velocity[1][b[0], b[1], b[2]] = coef[0] * self.velocity[1][b[0], b[1], b[2]]
            self.velocity[2][b[0], b[1], b[2]] = coef[0] * self.velocity[2][b[0], b[1], b[2]]


            self.velocity[0][s[0], s[1], s[2]] = coef[0] * self.velocity[0][s[0], s[1], s[2]]
            self.velocity[1][s[0], s[1], s[2]] = coef[0] * self.velocity[1][s[0], s[1], s[2]]
            self.velocity[2][s[0], s[1], s[2]] = coef[0] * self.velocity[2][s[0], s[1], s[2]]


            self.velocity[0][p[0], p[1], p[2]] = coef[1] * self.velocity[0][p[0], p[1], p[2]]
            self.velocity[1][p[0], p[1], p[2]] = coef[1] * self.velocity[1][p[0], p[1], p[2]]
            self.velocity[2][p[0], p[1], p[2]] = coef[1] * self.velocity[2][p[0], p[1], p[2]]

        #-------- vorticity penalization (curl on grid) :
#        self.vorticity.data = DifferentialOperator(self.velocity, self.vorticity, choice='curl', topology=self.velocity.topology).apply()


        self.compute_time = time.time() - self.compute_time
        self.total_time += self.compute_time

        return self.compute_time

    def printComputeTime(self):
#        self.timings_info[0] = "\"Penalization total\""
#        self.timings_info[1] = str(self.total_time)
        print "Penalization total time : ", self.total_time
        print "Time of the last Penalization iteration :", self.compute_time

    def __str__(self):
        s = "Penalization_d (DiscreteOperator). " + DiscreteOperator.__str__(self)
        return s

if __name__ == "__main__":
    print __doc__
    print "- Provided class : Penalization_d"
    print Penalization_d.__doc__
