# -*- coding: utf-8 -*-
"""
@package parmepy.operator Poisson solver
Continous Poisson solver operator (F2PY)
"""
from parmepy.operator.continuous import Operator
from parmepy.f2py import fftw2py
from parmepy.mpi.topology import Cartesian
from parmepy.operator.poisson_fft import PoissonFFT
from parmepy.constants import debug


class Poisson(Operator):
    """
    Poisson solver operator representation using FFT
    """

    @debug
    def __init__(self, velocity=None, vorticity=None,
                 resolutions=None, method='',
                 **other_config):
        """
        Constructor.
        Create a Poisson solver operator using FFT.

        @param velocity ContinuousVectorField : velocity variable.
        @param vorticity ContinuousVectorField : vorticity variable.
        """
        Operator.__init__(self, [velocity, vorticity])
        self.velocity = velocity
        self.vorticity = vorticity
        self.resolutions = resolutions
        self.method = method
        self.config = other_config

    @debug
    def setUp(self):
        """
        Poisson solver operator discretization method.
        Create a discrete Poisson solver operator from given specifications.

        @param idVelocityD : Index of velocity discretisation to update from vorticity.
        @param idVelocityD : Index of vorticity.
        @param topology : topology of the input fields
        """
        Operator.setUp(self)
        print self.method
        localres, localoffset = fftw2py.init_fftw_solver(
                                    self.resolutions[self.vorticity],
                                    self.domain.length)
        print 'localres, localoffset', localres, localoffset
        topodims = self.resolutions[self.vorticity] / localres
        print 'topodims POISSON', topodims
        #variables discretization
        for v in self.variables:
            topoId, topo = self.domain.addTopology(
                Cartesian.withResolution(
                    domain=self.domain, topoResolution=topodims,
                    globalMeshResolution=self.resolutions[v]))
            df, dfId = v.discretize(topo)
            self.discreteFieldId[v] = dfId
        self.discreteOperator = PoissonFFT(self, method=self.method,
                                           **self.config)
        self.discreteOperator.setUp()

    def __str__(self):
        """ToString method"""
        s = " Poisson solver operator (ContinuousOperator)"
        if self.discreteOperator is not None:
            s += "with the following discretization:\n"
            s += str(self.discreteOperator)
        else:
            s += ". Not discretised"
        return s + "\n"

if __name__ == "__main__":
    print __doc__
    print "- Provided class : Poisson"
    print Poisson.__doc__
