# -*- coding: utf-8 -*-
"""
@package operator
Discrete Poisson solver operator using FFT
"""
from parmepy.f2py import fftw2py
from parmepy.operator.discrete import DiscreteOperator
from parmepy.constants import np, ORDER, PARMES_REAL, debug
import time


class PoissonFFT(DiscreteOperator):
    """
    Operator representation.
    """

    @debug
    def __init__(self, poisson, method=''):
        """
        Constructor.
        @param velocity : descretized velocity to update from vorticity.
        @param vorticity
        """
        DiscreteOperator.__init__(self, poisson)
        ## Parent Continous Op.
        self.poisson = poisson
        ## Velocity.
        self.velocity = self.poisson.velocity.discreteField[
            self.poisson.discreteFieldId[self.poisson.velocity]]
        ## Vorticity.
        self.vorticity = self.poisson.vorticity.discreteField[
            self.poisson.discreteFieldId[self.poisson.vorticity]]
        self.method = method
        self.domain = self.poisson.domain
        self.compute_time = 0.

    @debug
    def setUp(self):
        # topology is recovered from discrete velocity field topology
        self.topology = self.velocity.topology
        self.coords = self.topology.mesh.coords
        self.step = self.topology.mesh.space_step
        self.local_start = self.topology.mesh.local_start
        self.local_end = self.topology.mesh.local_end

    @debug
    def apply(self, t, dt, ite):
        """
        Apply operator.
        """
        self.compute_time = time.time()

        #=== RECOVER VELOCITY FIELD FROM VORTICITY FIELD ===
        if (self.domain.dimension == 2):
            self.velocity.data[0],\
            self.velocity.data[1] =\
                fftw2py.solve_poisson_2d(self.vorticity.data,
                                        self.velocity.data[0],
                                        self.velocity.data[1])

        elif (self.domain.dimension == 3):
            self.velocity.data[0],\
            self.velocity.data[1],\
            self.velocity.data[2] = \
                fftw2py.solve_poisson_3d(self.vorticity.data[0],
                                        self.vorticity.data[1],
                                        self.vorticity.data[2],
                                        self.velocity.data[0],
                                        self.velocity.data[1],
                                        self.velocity.data[2])

        else:
            raise ValueError("invalid problem dimension")


        #============== VELOCITY CORRECTION ================

#        surf = (self.domain.length[1]-self.domain.origin[1]) * \
#               (self.domain.length[2]-self.domain.origin[2])

        #============= CORRECTION VEL X : u_x ==============

        # computation of the current flow rate: int_y int_z uX dydz
#        debX = 0.
#        debX = np.sum(self.velocity.data[0][0, :, :]) 
#        debX = debX * self.step[1] * self.step[2]

#        self.velocity.data[0] = \
#            self.velocity.data[0] + 1. - debX / surf

#        #============= CORRECTION VEL Y : u_y ==============

#        # computation of the current flow rate: int_y int_z uY dydz
#        debY = 0.
#        debY = np.sum(self.velocity.data[1][0, :, :]) 
#        debY = debY * self.step[1] * self.step[2]

#        # computation of omega_z circulation
#        omZbar = 0.
#        omZbar = np.sum(self.vorticity.data[2])
#        omZbar = omZbar * self.topology.mesh.space_step_size[1] * \
#                 self.topology.mesh.space_step_size[2] / surf

#        self.velocity.data[1] = \
#            self.velocity.data[1] \
#            + omZbar * self.coords[0][self.local_start:self.local_end] \
#            - omZbar * self.coords[0][self.local_start] \
#            - debY / surf

#        #============= CORRECTION VEL Z : u_z ==============

#        # computation of the current flow rate: int_y int_z uZ dydz
#        debZ = 0.
#        debZ = np.sum(self.velocity.data[2][0, :, :]) 
#        debZ = debZ * self.step[1] * self.step[2]

#        # computation of omega_y circulation
#        omYbar = 0.
#        omYbar = np.sum(self.vorticity.data[1])
#        omYbar = omYbar *self.step[1] * self.step[2] / surf

#        self.velocity.data[2] = \
#            self.velocity.data[2] \
#            - omYbar * self.coords[0][self.local_start:self.local_end] \
#            + omYbar * self.coords[0][self.local_start] \
#            - debZ / surf

        # PENSER AU MPI_REDUCE !!


        self.compute_time = time.time() - self.compute_time
        self.total_time += self.compute_time
        return self.compute_time

    def printComputeTime(self):
#        self.timings_info[0] = "\"Poisson solver total\""
#        self.timings_info[1] = str(self.total_time)
        print "Poisson total time : ", self.total_time
        print "Time of the last Poisson solver loop :", self.compute_time

    def __str__(self):
        s = "Poisson_d (DiscreteOperator). " + DiscreteOperator.__str__(self)
        return s

if __name__ == "__main__":
    print __doc__
    print "- Provided class : Poisson_d"
    print Poisson_d.__doc__
