"""
@package parmepy.operator.remeshing

Discrete remeshing representation
"""
from parmepy.operator.discrete import DiscreteOperator
from parmepy.constants import np
from parmepy.gpu import PARMES_REAL_GPU
import time


class Remeshing(DiscreteOperator):
    """
    Remeshing operator representation.
    """

    def __init__(self, partPositions, partScalar, resscal, method=None):
        """
        Create a Remeshing operator.

        @param partPositions : particles positions.
        @param partScalar : particles scalar.
        @param resscal : result grid scalar values.
        @param method : the method to use.
        """
        DiscreteOperator.__init__(self)
        ## Particles positions.
        self.ppos = partPositions
        ## Particles scalar values.
        self.pscal = partScalar
        ## Result scalar
        self.res_scalar = resscal
        self.addVariable([self.ppos, self.pscal, self.res_scalar])
        self.input = [self.ppos, self.pscal]
        self.output = [self.res_scalar]
        self.method = method
        ## Compute time detailed per directions
        self.compute_time = [0., 0., 0.]
        self.name = "remeshing"
        self.call_number = [0, 0, 0]
        self.gpu_precision = PARMES_REAL_GPU

    def apply(self, t, dt, splittingDirection):
        """
        Apply Remeshing operator.

        @param t : current time.
        @param dt : time step.
        @param splittingDirection : Direction of splitting.

        Remeshing algorithm:
        @li 1. Remesh particles on grid
                 - Use a M'6 formula
        @li 2. Profile timings of OpenCL kernels.
        """
        c_time = 0.
        self.call_number[splittingDirection] += 1
        if self.numMethod is not None:
            ## Launching kernel
            evt = self.numMethod.launch(self.ppos.gpu_data,
                                        self.pscal.gpu_data,
                                        self.res_scalar.gpu_data,
                                        self.gpu_precision(self.res_scalar.topology.mesh.origin[splittingDirection]),
                                        self.gpu_precision(self.res_scalar.topology.mesh.size[splittingDirection]))
            for df in self.output:
                df.contains_data = False
            # Get timpings from OpenCL events
            self.numMethod.finish()
            c_time += (evt.profile.end - evt.profile.start) * 1e-9
            self.compute_time[splittingDirection] += c_time
            self.total_time += c_time
        return c_time

    def printComputeTime(self):
        self.timings_info[0] = "\"Remeshing total\" \"Remeshing x\" \"Remeshing y\" \"Remeshing z\" "
        self.timings_info[1] = str(self.total_time) + " " + str(self.compute_time[0]) + " "
        self.timings_info[1] += str(self.compute_time[1]) + " " + str(self.compute_time[2]) + " "
        print "Remeshing total time : ", self.total_time, self.call_number
        print "\t Remeshing: ", self.compute_time, self.numMethod.call_number

    def __str__(self):
        s = "Remeshing (DiscreteOperator). " + DiscreteOperator.__str__(self)
        return s

if __name__ == "__main__":
    print __doc__
    print "- Provided class : RemeshingDOp"
    print RemeshingDOp.__doc__
