# -*- coding: utf-8 -*-
"""
@package parmepy.operator.stretching

Stretching operator representation


"""
from parmepy.constants import debug
from parmepy.operator.continuous import Operator
from parmepy.mpi.topology import Cartesian
from parmepy.operator.stretching_d import Stretching_d


class Stretching(Operator):
    """
    Stretching operator representation
    """

    @debug
    def __init__(self, velocity, vorticity,
                 resolutions=None, method='FD_order4 RK3',
                 **other_config):
        """
        Constructor.
        Create a Stretching operator from given
        velocity and vorticity variables.

        @param velocity ContinuousVectorField : velocity variable.
        @param vorticity ContinuousVectorField : vorticity variable.
        """
        Operator.__init__(self, [velocity, vorticity])
        ## velocity variable (vector)
        self.velocity = velocity
        ## vorticity variable (vector)
        self.vorticity = vorticity
        self.resolutions = resolutions
        self.method = method
        self.config = other_config

    @debug
    def setUp(self):
        """
        Stretching operator discretization method.
        Create a discrete Stretching operator from given specifications.

        @param idVelocityD : Index of velocity discretisation to use.
        @param idVorticityD : Index of vorticity discretisation to use.
        @param propertyOp : property garantied by the stretching Op
            (divConservation, massConservation) default : divConservation
        @param method : the method to use :
            space : (FD_order2, FD_order4, ..) default : FD_order4
            and time : (Euler, RK, ..) default : RK3
        """
        Operator.setUp(self)
        print self.method
        nbGhosts = 0
        if self.method.find('FD_order2') >= 0:
            nbGhosts = 1
        elif self.method.find('FD_order4') >= 0:
            nbGhosts = 2
        else:
            nbGhosts = 2
        #variables discretization
        for v in self.variables:
            topoId, topo = self.domain.addTopology(
                Cartesian(domain=self.domain, dim=self.domain.dimension,
                          globalMeshResolution=self.resolutions[v],
                          ghosts=[nbGhosts, nbGhosts, nbGhosts]))
            df, dfId = v.discretize(topo)
            self.discreteFieldId[v] = dfId
        self.discreteOperator = Stretching_d(self, method=self.method,
                                             **self.config)
        self.discreteOperator.setUp()

    def __str__(self):
        """ToString method"""
        s = " Stretching operator (ContinuousOperator)"
        if self.discreteOperator is not None:
            s += "with the following discretization:\n"
            s += str(self.discreteOperator)
        else:
            s += ". Not discretised"
        return s + "\n"

if __name__ == "__main__":
    print __doc__
    print "- Provided class : Stretching"
    print Stretching.__doc__
