# -*- coding: utf-8 -*-
"""
@package parmepy.operator.stretching_d

Discrete stretching representation
"""

from parmepy.constants import debug
from numpy import linalg as LA
from parmepy.operator.discrete import DiscreteOperator
from parmepy.numerics.differentialOperator import DifferentialOperator
from parmepy.numerics.fct2op import Fct2Op
from parmepy.numerics.integrators.euler import Euler
from parmepy.numerics.integrators.runge_kutta2 import RK2
from parmepy.numerics.integrators.runge_kutta3 import RK3
from parmepy.numerics.integrators.runge_kutta4 import RK4
import time


class Stretching_d(DiscreteOperator):
    """
    Stretching operator representation.
    DiscreteOperator.DiscreteOperator specialization.
    """

    @debug
    def __init__(self, stretch, method='FD_order4 RK3', propertyOp='divConservation'):
        """
        Constructor.
        Create a Stretching operator on a given continuous domain.
        Work on a given velocity and vorticity to compute the stretching term.

        @param stretch : Stretching operator.
        @param method : Method to use
        @param propertyOp : Operator property to satisfy (divConservation or massConservation)
        """
        DiscreteOperator.__init__(self, stretch)
        self.stretch = stretch
        ## Velocity.
        self.velocity = self.stretch.velocity.discreteField[
            self.stretch.discreteFieldId[self.stretch.velocity]]
        ## Vorticity.
        self.vorticity = self.stretch.vorticity.discreteField[
            self.stretch.discreteFieldId[self.stretch.vorticity]]
#        ## input fields
#        self.input = [self.velocity, self.vorticity]
        self.propertyOp = propertyOp
        self.method = method

    @debug
    def setUp(self):
        self.cststretch = 0.
        if self.method.find('Euler') >= 0:
            self.num_method = Euler
            self.cststretch = 2.0
        elif self.method.find('RK2') >= 0:
            self.num_method = RK2
            self.cststretch = 2.0
        elif self.method.find('RK3') >= 0:
            self.num_method = RK3
            self.cststretch = 2.5127
        elif self.method.find('RK4') >= 0:
            self.num_method = RK4
            self.cststretch = 2.7853
        else:
            self.num_method = RK3
            self.cststretch = 2.5127

    @debug
    def apply(self, t, dt, ite):
        """
        Apply Stretching operator.

        @param t : current time.
        @param dt : time step.
        @param ite : current iteration.

        Stretching algorithm:
        @li 1. discretization of the divergence operator (div wu)
        or w grad(u) : \n
        @li 2. Time integration. Performs a method choose to resolove
        the equation dw/dt = div(wu) or dw/dt = w grad(u).
         -define the function of ODEsolver as the result of the first step
         -integrate with the chosen method (RK4. RK2. euler)
        """
        self.compute_time = time.time()

        if self.num_method is not None:

            if (self.propertyOp.find('divConservation') < 0
                and self.propertyOp.find('massConservation') < 0):
                print 'No or wrong stretching operator property specified, use default divConservation'
                self.propertyOp = 'divConservation'

            if self.propertyOp == 'massConservation':
                ## Determination of Stretching time step (stability condition)
                self.ndt = 1
                self.dtstretch = dt

                ## fct2Op
                self.function = Fct2Op(self.vorticity,
                                       self.velocity,
                                       choice='divWU',
                                       topology=self.velocity.topology)

            else: # self.propertyOp == 'divConservation'

                ## Space discretization of grad(u)
                self.stretchOp = DifferentialOperator(
                    self.vorticity, self.velocity,
                    method='FD_order4',
                    choice='gradV')

                gradResult, maxgersh = self.stretchOp()

                print "Maxgersh", maxgersh
                print "dtAdvec", 0.5 / maxgersh[1]

                ## Determination of Stretching time step (stability condition)
                self.ndt = self.stabilityTest(dt, maxgersh[0], self.num_method, self.cststretch)
                self.dtstretch = dt / float(self.ndt)

                ## fct2Op
                self.function = Fct2Op(self.vorticity.data,
                                       gradResult,
                                       method='FD_order4',
                                       choice='gradV',
                                       topology=self.vorticity.topology)

            ## Time integration
            for i in range(self.ndt):
                methodInt = self.num_method(self.function)
                self.vorticity.data = methodInt(self.function.apply,
                                                t, self.dtstretch,
                                                self.vorticity.data)
#            print 'norm omega', LA.norm(self.vorticity.data)

            self.compute_time = time.time() - self.compute_time
            self.total_time += self.compute_time

        return self.compute_time

    def stabilityTest(self, dt, maxi, num_method, cststretch):
        dtstretch = dt
        dtstretch = min(dtstretch, cststretch / maxi)
        if dt == dtstretch:
            print "dt, ndt , dtstretch, upperbound", dt, int(dt / dtstretch),\
                  dt / float(int(dt / dtstretch)), cststretch / maxi
            return int(dt / dtstretch)
        else:
            print "dt, ndt , dtstretch, upperbound", dt, int(dt / dtstretch)\
                  + 1, dt / float(int(dt / dtstretch) + 1), cststretch / maxi
            return int(dt / dtstretch) + 1

    def printComputeTime(self):
#        self.timings_info[0] = "\"Stretching total\""
#        self.timings_info[1] = str(self.total_time)
        print "Stretching total time : ", self.total_time
        print "Time of the last Stretching iteration :", self.compute_time

    def __str__(self):
        s = "Stretching_d (DiscreteOperator). " +\
            DiscreteOperator.__str__(self)
        return s

if __name__ == "__main__":
    print __doc__
    print "- Provided class : Stretching_d"
    print Stretching_d.__doc__
