"""
@package parmepy.operator.transport_d

Discrete transport representation
"""
from parmepy.constants import np
from parmepy.operator.discrete import DiscreteOperator
from parmepy.gpu import PARMES_REAL_GPU
import time


class Transport_d(DiscreteOperator):
    """
    Particle transport operator representation.

    """

    def __init__(self, advec, idVelocityD=0, idScalarD=0,
                 result_position=None, result_scalar=None, method=None):
        """
        Create a Advection operator.
        Work on a given scalar at a given velocity to produce scalar
        distribution at new positions.

        @param advec : Transport operator
        @param idVelocityD : Index of velocity discretisation to use.
        @param idScalarD : Index of scalar discretisation to use.
        @param result_position : result position.
        @param result_scalar : result scalar.
        @param method : the method to use.
        """
        DiscreteOperator.__init__(self)
        ## Velocity.
        self.velocity = advec.velocity.discreteField[idVelocityD]
        ## Transported scalar.
        self.scalar = advec.scalar.discreteField[idScalarD]
        ## Result position
        self.res_position = result_position.discreteField[0]
        ## Result scalar
        self.res_scalar = result_scalar.discreteField[0]
        self.input = [self.velocity, self.scalar]
        self.output = [self.res_position, self.res_scalar]
        self.method = method
        ## Previous splitting direction
        self.old_splitting_direction = None
        ## Compute time detailed per directions
        self.compute_time = [0., 0., 0.]
        ## Compute time for copy detailed per directions
        self.compute_time_copy = [0., 0., 0.]
        ## Compute time for transposition detailed per directions
        self.compute_time_swap = [0., 0., 0.]
        self.name = "advection"
        self.call_number = [0, 0, 0]
        self.gpu_precision = PARMES_REAL_GPU

    def apply(self, t, dt, splittingDirection):
        """
        Apply advection operator.

        @param t : current time.
        @param dt : time step.
        @param splittingDirection : Direction of splitting.

        Advection algorithm:
        @li 1. Particle initialization : \n
                 - by copy scalar from grid to particles if previous splitting
                 direction equals current splitting direction.\n
                 - by transposition of scalar from grid to particle.
        @li 2. Particle advection :\n
                 - compute particle position in splitting direction as a
                 scalar. Performs a RK2 resolution of dx_p/dt = a_p.
        @li 3. Profile timings of OpenCL kernels.
        """
        c_time, c_time_init = 0., 0.
        self.call_number[splittingDirection] += 1
        if self.numMethod is not None and self.init_transpose is not None and self.init_copy is not None:
            # Particle init
            if (self.old_splitting_direction == splittingDirection) or self.old_splitting_direction is None:
                evt_init = self.init_copy.launch(self.scalar.gpu_data,
                                                 self.res_scalar.gpu_data)
            else:
                if self.scalar.topology.dim == 2:
                    evt_init = self.init_transpose.launch(self.scalar.gpu_data,
                                                          self.res_scalar.gpu_data)
                else:
                    if min(self.old_splitting_direction, splittingDirection) == 0:
                        evt_init = self.init_transpose.launch(0,
                                                              self.scalar.gpu_data,
                                                              self.res_scalar.gpu_data)
                    else:
                        evt_init = self.init_transpose.launch(1,
                                                              self.scalar.gpu_data,
                                                              self.res_scalar.gpu_data)
            # Advection
            evt = self.numMethod.launch(self.velocity.gpu_data[splittingDirection],
                                        self.res_position.gpu_data,
                                        self.gpu_precision(dt),
                                        self.gpu_precision(self.scalar.topology.mesh.origin[splittingDirection]),
                                        self.gpu_precision(self.scalar.topology.mesh.size[splittingDirection]))
            for df in self.output:
                df.contains_data = False
            # Get timpings from OpenCL events
            self.numMethod.finish()
            c_time_init = (evt_init.profile.end - evt_init.profile.start) * 1e-9
            c_time = (evt.profile.end - evt.profile.start) * 1e-9
            self.compute_time[splittingDirection] += c_time
            if (self.old_splitting_direction == splittingDirection) or self.old_splitting_direction is None:
                self.compute_time_copy[splittingDirection] += c_time_init
            else:
                self.compute_time_swap[min(self.old_splitting_direction, splittingDirection)] += c_time_init
            self.total_time += (c_time + c_time_init)
            self.old_splitting_direction = splittingDirection
        return (c_time + c_time_init)

    def printComputeTime(self):
        self.timings_info[0] = "\"Advection total\" \"copy\" \"swap xy\" \"swap xz\" \"Advection x\" \"Advection y\" \"Advection z\" "
        self.timings_info[1] = str(self.total_time) + "  " + str(self.compute_time_copy[0]) + "  "
        self.timings_info[1] += str(self.compute_time_swap[0]) + "  " + str(self.compute_time_swap[1]) + "  "
        self.timings_info[1] += str(self.compute_time[0]) + "  " + str(self.compute_time[1]) + "  " + str(self.compute_time[2]) + "  "
        print "Advection total time : ", self.total_time, self.call_number
        print "\t Advection init copy :", self.compute_time_copy, self.init_copy.call_number
        print "\t Advection init swap :", self.compute_time_swap, self.init_transpose.call_number
        print "\t Advection :", self.compute_time, self.numMethod.call_number

    def __str__(self):
        s = "Advection_P (DiscreteOperator). " + DiscreteOperator.__str__(self)
        return s

if __name__ == "__main__":
    print __doc__
    print "- Provided class : Transport_d"
    print AdvectionDOp.__doc__
