"""
@package parmepy.problem.problem

Complete problem description.
"""
from abc import ABCMeta, abstractmethod
from parmepy.constants import debug
from parmepy import __VERBOSE__
from parmepy.operator.monitors.monitoring import Monitoring


class Problem(object):
    """
    Problem representation.

    Contains several operators that apply on variables.
    Variables are defined on different domains.\n
    The problem contains a solver that define how the problem is solved.\n
    A problem is based on a topology that deploys
    its resolution over MPI processus.\n
    Finally, the problem handle outputs thanks to a io object.
    """

    __metaclass__ = ABCMeta

    @debug
    def __new__(cls, *args, **kw):
        return object.__new__(cls, *args, **kw)

    @debug
    @abstractmethod
    def __init__(self, operators, monitors=[]):
        """
        Create a transport problem instance.

        @param operators : list of operators.
        """
        ## Problem operators
        self.operators = operators
        for m in monitors:
            self.operators.append(m)
        ## Computes time step and manage iterations
        self.timer = None
        ## Timings informations
        self.time_info = ["", ""]
        self.domain = self.operators[0].variables[0].domain
        for op in self.operators:
            for v in op.variables:
                if not self.domain is v.domain:
                    raise ValueError("Problem must have only one " +
                                     "domain for variables.")

    @debug
    def setUp(self, t_end, dt):
        """
        Set solver.
        @param t_end : duration of the simulation.
        @param dt : simulation time step.
        """
        [op.setUp() for op in self.operators]
        self.timer = Timer(t_end, dt)
        if  __VERBOSE__:
            print "==== Variables initialization ===="
        for op in self.operators:
            for v in op.variables:
                v.initialize()
        if __VERBOSE__:
            print "===="

    @debug
    def solve(self):
        """
        Solve problem.

        Performs simulations iterations by calling each
        operators of the list until timer ends.\n
        At end of time step, call an io step.\n
        Displays timings at simulation end.
        """
        ## Iterations counter
        ite = 0
        print "\n\n Start solving ..."
        for op in self.operators:
            if isinstance(op, Monitoring):
                op.apply(self.timer.t, self.timer.dt, ite)
        while not self.timer.end:
            ite = ite + 1
            if (self.domain.topologies[0].rank == 0):
                print "==== Iteration : {0:3d}   t={1:6.3f} ====".format(
                    self.timer.ite, self.timer.t + self.timer.dt)
            for op in self.operators:
                op.apply(self.timer.t, self.timer.dt, ite)
            self.timer.step()
        print "\n\n End solving\n"
        print "==== Finalization ===="
        for op in self.operators:
            op.finalize()
        print "===\n"
        print "=== Timings ==="
        if (self.domain.topologies[0].rank == 0):
            for op in self.operators:
                op.printComputeTime()
                self.time_info[0] += op.time_info[0]
                self.time_info[1] += op.time_info[1]
            print "\n"
            print "===\n"

    def __str__(self):
        """ToString method"""
        s = "Problem based on\n"
        s += str(self.domain.topologies)
        s += "with following operators : \n"
        for op in self.operators:
            s += str(op)
        s += "working on fields : \n"
        for f in self.variables:
            s += str(f)
        s += "defined on domains : \n"
        for d in self.domains:
            s += str(d)
        s += "\nsolved by : \n" + str(self.solver)

        return s


class Timer:
    """
    Manage time steps and simulation end.

    """

    def __init__(self, t_end, dt, t_init=0.):
        """
        Creates a Timer.

        @param t_end : Simulation final time.
        @param dt : Time step.
        @param t_init : Simulation starting time.
        """
        ## Simulation final time
        self.t_end = t_end
        ## Simulation time step
        self.dt = dt
        ## Simulation current time
        self.t = t_init
        ## Is simulation is terminated
        self.end = False
        ## Iteration counter
        self.ite = 0

    def step(self):
        """
        Update current time.
        """
        if self.t < self.t_end:
            self.ite += 1
            self.t += self.dt
            if self.t >= self.t_end:
                self.end = True

    def __str__(self):
        s = "Timer (DiscreteOperator). "
        return s

if __name__ == "__main__":
    print __doc__
    print "- Provided class : problem, Timer"
    print Problem.__doc__
    print Timer.__doc__
